#pragma warning(disable: 4100 4201 4514)

#include <windows.h>
#include <stdio.h>

#include "Decode.h"
#include "VidCap.h"
#include "resource.h"


/*
 * Ein wenig Theorie
 *
 * Bei Syster/Nagravision werden die Positionen der Zeilen permutiert. Der Decoder hat RAM fr 32
 * Zeilen und da schreibt er auch erstmal die letzten 32 Zeilen eines verschlsselten Halbbildes
 * rein bevor irgendwas anderes passiert. Das sind nmlich die ersten 32 Zeilen des decodierten
 * Halbbildes, was man auch sehr schn erkennt, wenn man sich das verschlsselte Bild mal anschaut.
 * Danach schiebt er Zeile fr Zeile an den Fernseher und holt sich fr jede Zeile die er
 * rausschiebt sofort eine neue. Die Reihenfolge bestimmt, wie bei videocrypt, ein pseudo random
 * number generator, der durch einen verschlsselten seed gestartet wird. Dieser prng ist ein sehr
 * einfaches Modell. Er besteht aus einer Liste von 256 geschickt angeordneten Zahlen zwischen 0
 * und 31. Der seed besteht aus zwei Byte, das erste gibt an, an welcher Stelle in dieser Liste
 * begonnen wird, das zweite, mit welcher ungeraden Schrittweite round robin durch die Liste
 * gegangen wird (deshalb auch 256 x 128 = 32768 mgliche Permutationen). Heraus kommt jeweils
 * eine Zahl zwischen 0 und 31. Dies ist der Index der Zeile im RAM des Decoders, die als nchstes
 * an den Ferseher geschoben wird.
 * 
 * Der Prozessor und die gesammte restliche Hardware wird duch den Zeilentakt des Videosignals
 * getaktet, damit eine hundertprozentige Synkronitt gewhrleistet werden kann. 
 * 
 * Die Decodierungssoftware fr den PC funktioniert nun so:
 * Die Software braucht als Grundlage die oben beschriebene Liste. Es handelt sich einfach nur um
 * eine Datei, die 256 Zeilen enthlt, jede Zeile mit einer Zahl zwischen 0 und 31. Daraus werden
 * nun alle! 32768 Mglichen Permutationen errechnet.
 * 
 * Die Arbeit des Algorithmus besteht nun darin, so schnell wie mglich die richtigen zu finden.
 * Der Algorithmus beruht auf folgendem Prinzip: "Wenn zwei im unverschlsselten Bild
 * aufeinanderfolgende Bildzeilen eines Fields (Halbbildes), die im verschlsselten Bild von
 * einander entfernt liegen, nach einer Transformation mit einer bestimmten Permutation wieder
 * korrekterweise untereinander liegen, dann ist dieser Schlssel mit groer Wahrscheinlichkeit
 * der richtige." Selbstverstndlich sind zwei Zeilen nicht ausreichend um sicher zu sein, das
 * die angenommene Permutation auch die richtige ist. Im Folgenden werde ich diese Zeilenpaare
 * "Kandidaten" nennen. 
 * 
 * Aber wie kann man herausfinden, ob zwei Zeilen zusammen gehren ? Indem man beispielsweise
 * ihre Korrelationsrate berechnet: Wenn die Summe der Betrge der Luminanzdifferenz jedes Pixels
 * der beiden Zeilen "klein" ist, dann liegen sie wahrscheinlich nahe bei einander. Der Erfolg
 * dieser Methode hngt davon ab, da weit voneinander entfernte Zeilen des Bildes relativ
 * unterschiedlich, nahe beieinanderligende aber sehr hnlich sind. Das ist bei Video im 
 * allgemeinen der Fall, auer wenn das Bild sehr einfrmig ist (z.B. ein einfarbiger Hintergrund)
 * oder wenn vertikale Symmetrien vorhanden sind (ein einfarbiges, geometrisches Bild mit
 * senkrechtem Rand). Kurz, je inhaltsreicher das Bild ist, um so besser geht es. Um die
 * Gltigkeit einer Permutation festzustellen, mu man also
 * 
 * 1.) eine grere Anzahl Kandidaten whlen. Je kleiner die Anzahl, um so schneller aber
 *     auch um so unzuverlssiger ist der Agorithmus.
 * 
 * 2.) denjenigen variablen Schlssel bestimmen, der das beste Korrelationsergebnis aller
 *     seiner Kandidaten erhlt (es gengt, die individuellen Korrelationsraten zu summieren).
 * 
 * Die Anzahl der Pixel, die zur Berechnung der Korrelationsrate verwendet werden, kann geringer
 * sein als die Anzahl der Pixel einer Zeile. Dabei ist es sinnvoll, aus verschiedenen Bereichen
 * der Zeile einige nebeneinanderliegende Pixel zu nehmen, um Kontraste in der Senkrechten
 * gnstig auszuwerten. Je mehr Pixel man nimmt, um so zuverlssiger, aber auch um so langsamer
 * arbeitet der Algorithmus. Alle diese Parameter haben eine groe Wirkung auf die Geschwindigkeit
 * des Algorithmus. Um eine Vorstellung davon zu geben: Wir schauen uns 12 Zeilen an. Daraus
 * ergeben sich 800 mglichen Kandidaten (das sind alle mglichen Reihenfolgen dieser 12 Zeilen).
 * In jeder Zeile schauen wir uns jeweils 4 von 16 Pixeln an. D.h. es mssen fr jedes Bild 800
 * Korrelationsraten berechnet werden (800*2*(768*4/16))=307200 Pixel bearbeiten  und dann
 * 32768*12=393216 Mglichkeiten bewertet werden. Das mit 25 multipliziert ergibt den Aufwand
 * fr 1 Sekunde real-time Video. Hinzu kommt (fr 16 bit Farbe): 440k*25 = 11M Transferts vom
 * PCI-Bus zum Speicher durch den BT848 und 11 M Transferts vom Speicher zur Grafikkarte. Das
 * macht 22M/s. Ein AGP-Bus drfte die Last auf dem PCI-Bus halbieren. Die Last auf dem PCI-Bus
 * hat eine nicht vernachlssigbare Wirkung auf die Gte des Algorithmus, was teilweise den
 * Abfall der Bildrate zwischen den 8, 16 und 32-bit Versionen erklrt. 
 */

// Anzahl der Pixel die getestet werden
const int TESTPIXEL        = 24;

// Anzahl der Mglichkeiten
const int KOMBINATIONEN    = 32768;

// Kleinste Testzeile
const int KLEINSTE_ZEILE   = 96;

// Grte Testzeile
const int GROESSTE_ZEILE   = ZEILENZAHL-KLEINSTE_ZEILE-1;

// Name der Keydatei
const char KEY_DATEI[]     = "key.txt";

// Name der Indexdatei
const char INDEX_DATEI[]   = "MoreTV%02d.idx";

// Gre der YUV nach RGB Tabelle
const int YUV2RGB_SIZE     = 256*4*4; //64*64*64;

// Relative Position der Bits in der PalPlus Zeile fr eine horizontale Auflsung von 384
const short nPalPlusCode[] = { 9, 13, 18, 22, 27, 31, 36, 40, 45,  50,  55,  60,  67,  74};

//Tableau de Conversion
short YU12toRGB[256*256*256];

// Aus dieser Tabelle wird die Permutation errechnet
char	   pKey[256];

// Anzahl der Zeilen die getestet werden
int      nTestzeilen;

// Jede Mglichkeit der Zeilenpermutation
short   *pKombinationen;
int      nZeilenPaarKorrelation[INDEXSIZE];

CSyster::CSyster(int nTestzeilen2) 
{
	 nTestzeilen=nTestzeilen2;
   bError = false;

   pKombinationen = new short[KOMBINATIONEN*nTestzeilen2];
   if(pKombinationen == NULL)
   {
      bError = true;
      return;
   }

   // Permutationstabelle aus Datei einlesen
   if(LeseKeyDatei() == false)
   {
      MessageBoxID(IDS_ERR_READ_KEY, MB_ICONEXCLAMATION);
      bError = true;
      return;
   }

   // Wenn bereits ein Indexfile existiert, brauchen die Daten nicht mehr 
   // berechnet zu werden
   if(LeseIndexDatei() == true)
      return;

   // IndexTabellen aufbauen
   if(BaueIndex() == false)
   {
      bError = true;
      return;
   }

   // Das ganze frs nchste mal abspeichern
   if(SchreibeIndexDatei() == false)
   {
      bError = true;
      return;
   }
}


CSyster::~CSyster()
{
   if(pKombinationen != NULL)
      delete[] pKombinationen;
}


bool CSyster::FehlerStatus()
{
   return bError;
}


bool CSyster::BaueIndex()
{
   // Anzahl der Paare im Unverschlsseltem Zustand 
   // Das heist, die Zeilen sind direkt untereinander
	// GROESSTE_ZEILE => 190
	// KLEINSTE_ZEILE => 96
  // nAnzahlPaare => 94
   nAnzahlPaare = GROESSTE_ZEILE - KLEINSTE_ZEILE;

   nPaarNummer = new int[KOMBINATIONEN];
   if(nPaarNummer == NULL)
      return false;

   // In dieser Tabelle stehen alle Testzeilen einer Kombination hintereinander
   nPaarDanach = new int[KOMBINATIONEN*nAnzahlPaare];
   if(nPaarDanach == NULL)
      return false;

   nPaarDavor = new int[ZEILENZAHL*ZEILENZAHL];
   if(nPaarDavor == NULL)
      return false;

   nPaarTreffer = new int[ZEILENZAHL*ZEILENZAHL];
   if(nPaarTreffer == NULL)
      return false;

   int nZeilenIndex;

   // Kennzeichen fr nicht belegt
   for(nZeilenIndex=0; nZeilenIndex<ZEILENZAHL*ZEILENZAHL; nZeilenIndex++)
      nPaarDavor[nZeilenIndex] = -1;

   // Anzahl Treffer auf Null setzen
   for(nZeilenIndex=0; nZeilenIndex<ZEILENZAHL*ZEILENZAHL; nZeilenIndex++)
      nPaarTreffer[nZeilenIndex] = 0;

   // Eine doppelt verzeigerte Liste aller Zeilenpaare aufbauen
   int nPaarDavorIndex;
   int nPaarDanachIndex;
   int nKey;
   int pZeilen[ZEILENZAHL];
   int nZeileA;
   int nZeileB;

   // Schleife ber alle Kombination
   for(nKey=0; nKey<KOMBINATIONEN; 
	 nKey++)
   {
      // Zeilenfolge fr diese Kombination finden
      BerechnePermutation(nKey, pZeilen);

      // Erste Zeile eines Paars
      nZeileA = pZeilen[KLEINSTE_ZEILE];
      for(nZeilenIndex=KLEINSTE_ZEILE+1; nZeilenIndex<=GROESSTE_ZEILE; nZeilenIndex++)
      {
         // Zweite Zeile eines Paars
         nZeileB = pZeilen[nZeilenIndex];

         // Index immer aus "kleine Zeile * Zeilenzahl + groe Zeile" berechnen
         if(nZeileB < nZeileA)
            nPaarDavorIndex = nZeileB*ZEILENZAHL + nZeileA;
         else
            nPaarDavorIndex = nZeileA*ZEILENZAHL + nZeileB;

         // Index auf die Paar danach Tabelle
         nPaarDanachIndex = nKey*nAnzahlPaare + nZeilenIndex - (KLEINSTE_ZEILE+1);

         // Zeiger auf das Paar danach
         nPaarDanach[nPaarDanachIndex] = nPaarDavor[nPaarDavorIndex];
         
         // Zeiger auf das Paar davor
         nPaarDavor[nPaarDavorIndex] = nPaarDanachIndex;

         // Anzahl der Treffer von diesem Paars erhhen
         nPaarTreffer[nPaarDavorIndex]++;

         // Erste Zeile ist gleich zweiter Zeile
         nZeileA = nZeileB;
      }
   }

   for(nZeilenIndex=0; nZeilenIndex<KOMBINATIONEN; nZeilenIndex++)
      nPaarNummer[nZeilenIndex] = 0;

   int nCount = 0;
   int nTreffer;
   int nMaxTreffer;

   nPaarIndexAnzahl = 0;
   do
   {
      // Das Paar mit den meisten Treffern suchen
      nMaxTreffer = 0;
      nPaarDavorIndex = 0;
      for(nZeilenIndex=0; nZeilenIndex<ZEILENZAHL*ZEILENZAHL; nZeilenIndex++)
      {
         if(nPaarTreffer[nZeilenIndex] > nMaxTreffer)
         {
            nMaxTreffer = nPaarTreffer[nZeilenIndex];
            nPaarDavorIndex = nZeilenIndex;
         }
      }

      // Wenn das Paar benutzt wird dann merken
      nTreffer = PaarBewerten(nPaarDavorIndex);
      if(nTreffer > 0)
      {
         nCount += nTreffer;
         PaarMerken(nPaarDavorIndex);
      }

      nPaarTreffer[nPaarDavorIndex] = 0;
   }
   while(nCount < KOMBINATIONEN*nTestzeilen);

   delete[] nPaarNummer;
   delete[] nPaarTreffer;
   delete[] nPaarDavor;
   delete[] nPaarDanach;

   return true;
}


void CSyster::BerechnePermutation
(
   int   nKey,
   int  *pZeilen
)
{
   int   nStartwert = nKey>>7;
   int   nSchrittweite = (nKey&0x7f)*2+1;
   int   nZeile;
   int   nBuffer[32];/*={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,
					 17,18,19,20,21,22,23,24,25,26,27,28,29,30,31};// By Xetra*/
   //
   // Diese Funktion berechnet die Zeilenpermutation fr eine Kombination 
   // aus nStartwert und nSchrittweite.
   //
	 
   for(nZeile=0; nZeile<32; nZeile++)
      nBuffer[nZeile] = nZeile;

/*	 _asm {
		 pusha

	 ///////////////////////////////////////////////////////////////////////
	 // Cration d'un nouveau tableau Dbut
	 ///////////////////////////////////////////////////////////////////////
     lea edi, nBuffer
     xor eax, eax
     mov ecx, 32

   NouveauTableau:
		 stosd
		 inc eax
		 loop NouveauTableau

	 ///////////////////////////////////////////////////////////////////////
	 // Cration d'un nouveau tableau Fin
	 ///////////////////////////////////////////////////////////////////////
*/
	   // ZEILENZAHL  = 287 
	   for(nZeile=0; nZeile<ZEILENZAHL-32; nZeile++)
     {
       pZeilen[nZeile] = nBuffer[pKey[nStartwert]];
       nBuffer[pKey[nStartwert]] = nZeile+32;
       nStartwert += nSchrittweite;
       nStartwert &= 0xff;
     }
/*
     //287 - 32 = 255
		 mov ecx, 255
		 mov edx, 32
     mov edi, dword ptr pZeilen

   BouclenZeile:

     //  pZeilen[nZeile] = nBuffer[pKey[nStartwert]];
  		 lea esi, nBuffer
       // pKey = char [256];
  		 lea ebx, pKey
  		 mov eax, nStartwert
  		 xlat 
  		 shl eax, 2
  		 add esi, eax
  		 movsd

     //  nBuffer[pKey[nStartwert]] = nZeile+32;
  		 sub esi, 4
			 mov ebx, edi
			 mov edi, esi
			 mov eax, edx
		   stosd
			 mov edi, ebx
		   inc edx

     //  nStartwert += nSchrittweite;
     mov eax, nSchrittweite
		 add nStartwert, eax

     //  nStartwert &= 0xff;
		 and nStartwert, 0xFF

		 loop BouclenZeile

	 ///////////////////////////////////////////////////////////////////////
	 // Permutation Dbut
	 ///////////////////////////////////////////////////////////////////////
*/
     for(nZeile=0;nZeile<32;nZeile++)
        pZeilen[nZeile+ZEILENZAHL-32] = nBuffer[nZeile];
/*
	   mov edi, dword ptr pZeilen
		 lea esi, nBuffer
		 
     // (287 - 32) * 4 = 255 * 4 = 1020
	   add edi, 1020

     mov ecx, 32
		 rep movsd

	 ///////////////////////////////////////////////////////////////////////
	 // Permutation Dbut
	 ///////////////////////////////////////////////////////////////////////

		 popa
	 }*/
}

int CSyster::FindBlackLine // Auto Format by Xetra
(   int **pLinesAddress,
	int   nInputWidth
)
{
	const int MAX=50;   // Maximum average light of a "black pixel/line"
	int Lumiere=0;
	int nLigne=0;
	int nPixel=0;
	
	// *****Average light Line*****  
	while ((nLigne<ZEILENZAHL/2)&&(Lumiere<nInputWidth*MAX))
	{
		Lumiere=0;
		for(int nPixel=0; nPixel<nInputWidth; nPixel++)		
			Lumiere+=pLinesAddress[nLigne][nPixel] & 0xff;
		nLigne++;
	}
	
	// ******Max light Pixel*******  
	/*while ((pLinesAddress[nLigne][nPixel] & 0xff<MAX)&&(nLigne<ZEILENZAHL/2))
	{	
		if (++nPixel==nInputWidth)
		{	
			nPixel=0;
			nLigne++;
		}
	}*/
	return nLigne;
}

void CSyster::FindLineOrder
(
  // unsigned char **pLinesAddress, // Adresses => NULL
   unsigned char *pLinesAddress, // Adresses => NULL
   int  *pSortLines,    // Numero Ligne
   int   nInputWidth,   // Largeur des lignes => 352
   bool  bPalPlus       // Oui ou Non pour Pal => FALSE
)
{
   //
   // Die Daten kommen im YUY2 Format:
   //
   //  31--V0--24 23--Y1--16 15--U0--8 7--Y0--0
   //  MSB                                  LSB
   //
   // U0 und V0 gelten jeweils fr Y0 und Y1
   //

   int  nPalPlusZeile = -1;
   //int  nPixelSchritt = nInputWidth/TESTPIXEL/2; // 352 / 24 / 2 = 29.333 = 29
	 //int  nPixelSchritt = 8; // 352 / 24 / 2 = 29.333 = 29
	 int  nPixelSchritt = 14; // 352 / 24 / 2 = 29.333 = 29
	 //int  nPixelSchritt = 7; // 352 / 24 / 2 = 29.333 = 29
   //int  nStartPixel   = nPixelSchritt/2; // 29 / 2 = 14.5 = 14;
	 int  nStartPixel   = 7; // 29 / 2 = 14.5 = 14;
   int nTestPixel[ZEILENZAHL][TESTPIXEL]; // int nTestPixel[287][24];
   // Testpixel in einen Zwischenspeicher schreiben. Das erhht die Geschwindigkeit.
   
	 // int (32 bits)
	 // ZEILENZAHL => 287
	 // TESTPIXEL => 24
	 /*
   for(int nZeile=0; nZeile<ZEILENZAHL; nZeile++)
   {
      for(int nPixel=0; nPixel<TESTPIXEL; nPixel++)
         nTestPixel[nZeile][nPixel] = 
				 pLinesAddress[nZeile][nStartPixel+nPixel*nPixelSchritt] & 0xff;
   } */

   for(int nZeile=0; nZeile<ZEILENZAHL; nZeile++)
   {
       for(int nPixel=0; nPixel<TESTPIXEL; nPixel++)
         nTestPixel[nZeile][nPixel] = 
				 //pLinesAddress[nZeile][nStartPixel+nPixel*nPixelSchritt];
				 pLinesAddress[nStartPixel+(nPixel*nPixelSchritt)+(nZeile*352)];
	 			 //pLinesAddress[nZeile][nPixel] ;
   }

	 /*_asm {
     pusha

		 lea edi, nTestPixel
     mov esi, dword ptr pLinesAddress
		 mov esi, dword ptr [esi]
		 //lea esi, pLinesAddress

		 mov ecx, 287
   BoucleZEILENZAHL:
		 push ecx
     
     mov ecx, 24
   BoucleTestPixel:
		 lodsb
     and eax, 0xFF
		 stosd
		 loop BoucleTestPixel

		 //add esi, 80
		 add esi, 328
		 pop ecx
     loop BoucleZEILENZAHL

		 popa
	 }*/

   // PalPlus Zeile suchen
//   if(bPalPlus)
//      nPalPlusZeile = FindPalPlus(pLinesAddress, nInputWidth);

   // Die Korrelation zwischen den Zeilen berrechnen
   int nZeileA;
   int nZeileB;
   int nKorrelationZeile;
   int nKorrelationPixel;

   for(int nIndex=0; nIndex<nPaarIndexAnzahl; nIndex++)
   {
      nZeileA = nPaarIndex[nIndex] >> 9;
//			nZeileA /= 2;
      nZeileB = nPaarIndex[nIndex] & 0x1ff;
//			nZeileB /= 2;
      //nZeileA = nPaarIndex[nIndex] >> 10;
      //nZeileB = nPaarIndex[nIndex] & 0xff;

      nKorrelationZeile = 0;
      
      //for(int nPixel=nStartPixel;nPixel<nStartPixel+TESTPIXEL*nPixelSchritt;nPixel+=nPixelSchritt) // By Xetra
	  for(int nPixel=0; nPixel<TESTPIXEL; nPixel++)
      {
         //nKorrelationPixel = nTestPixel[nZeileA][nPixel] - nTestPixel[nZeileB][nPixel];
         nKorrelationPixel = nTestPixel[nZeileA][nPixel] - nTestPixel[nZeileB][nPixel];
         //nKorrelationPixel = (pLinesAddress[nZeileA][nPixel]&0xff) - (pLinesAddress[nZeileB][nPixel]& 0xff); // By Xetra
		 // Ein kleiner Ersatz fr die Funktion abs()
         //nKorrelationZeile += nKorrelationPixel;
		 
		 if(nKorrelationPixel < 0)
            nKorrelationZeile -= nKorrelationPixel;
         else
            nKorrelationZeile += nKorrelationPixel;
      }

      nZeilenPaarKorrelation[nIndex] = nKorrelationZeile;
   }

	 // Optimiz en Assembleur
   BerechnePermutation(SucheKey(nPalPlusZeile, bPalPlus), pSortLines);
	 //SucheKey(nPalPlusZeile, bPalPlus);
}


int CSyster::FindPalPlus
(
   int **pLinesAddress,
   int   nInputWidth
)
{
   int      nPixel;
   int      nZeile;
   int      nOffset;
   short   *pLine;

   // Die 32 mglichen Zeilen durchsuchen
   for(nZeile=0; nZeile<32; nZeile++)
   {  
      pLine = (short*)&pLinesAddress[nZeile][0];

      // Auf das Startbit warten
      nOffset = 0;
      for(nPixel=0; nPixel<10*nInputWidth/384; nPixel++)
      {
         // Startbit gefunden
         if(140 < (*(pLine+nOffset)&0xff))
            break;
         
         // Abstand Zeilenanfang bis Startbit
         nOffset++;
      }

      // Bitcode berprfen
      for(nPixel=0; nPixel<14; nPixel+=2)
      {
         // Fr andere horizontale Auflsungen als 384, Pixelabstand angleichen
         if(116 < (*(pLine+nOffset+nPalPlusCode[nPixel]*nInputWidth/384)&0xff))
            break;

         if(140 > (*(pLine+nOffset+nPalPlusCode[nPixel+1]*nInputWidth/384)&0xff))
            break;
      }

      // Wenn alle 7 Bit richtig sind dann wurde die Zeile gefunden
      if(nPixel == 14)
         return nZeile;
   }

   // Zeile nicht gefunden
   return -1;
}


bool CSyster::LeseIndexDatei()
{
   FILE *sFile;

   // Datei mit vorberechneten Tabellen einlesen
   char pDateiName[20];
   
   // Dateiname erzeugen
   sprintf(pDateiName, INDEX_DATEI, nTestzeilen);

   // Datei zum Lesen ffnen
   sFile = fopen(pDateiName, "rb");

   // Wenn die Datei nicht vorhanden ist, Tabellen neu berechnen
   if(sFile == NULL)
      return false;

   // Daten einlesen      
   if(fread(pKombinationen, sizeof(short), KOMBINATIONEN*nTestzeilen, sFile) != (size_t)(KOMBINATIONEN*nTestzeilen))
   {
      fclose(sFile);
      return false;
   }

   if(fread(nZeilenPaarKorrelation, sizeof(int), INDEXSIZE, sFile) != INDEXSIZE)
   {
      fclose(sFile);
      return false;
   }
      
   if(fread(nPaarIndex, sizeof(int), INDEXSIZE, sFile) != INDEXSIZE)
   {
      fclose(sFile);
      return false;
   }

   if(fread(&nPaarIndexAnzahl, sizeof(int), 1, sFile) != 1)
   {
      fclose(sFile);
      return false;
   }

   // Daten wurden erfolgreich eingelesen
   fclose(sFile);
   return true;
}


bool CSyster::LeseKeyDatei()
{
   // Permutationstabelle aus Datei einlesen
   int   nZlr;
   int   nInputValue;
   FILE *sFile = fopen(KEY_DATEI, "r");

   if(sFile == NULL)
      return false;

   for(nZlr=0; nZlr<256; nZlr++)
   {
      if(fscanf(sFile, "%d\n", &nInputValue) == EOF)
         return false;

      pKey[nZlr] = (char)(nInputValue&0x1f);
   }

   fclose(sFile);
   return true;
}


int CSyster::PaarBewerten
(
   int nPaarDavorIndex
)
{
   int nKey;
   int nTreffer = 0;
   int nNaechstesPaarIndex = nPaarDavor[nPaarDavorIndex];

   while(nNaechstesPaarIndex != -1)
   {
      nKey = nNaechstesPaarIndex/nAnzahlPaare;

      if(nPaarNummer[nKey] < nTestzeilen)
         nTreffer++;

      nNaechstesPaarIndex = nPaarDanach[nNaechstesPaarIndex];
   }

   return nTreffer;
}


void CSyster::PaarMerken
(
   int nPaarDavorIndex
)
{
   int nNaechstesPaarIndex;
   int nKey;
   int q;
   int nZeileA;
   int nZeileB;
   int nb;

   nZeileA = nPaarDavorIndex / ZEILENZAHL;
   nZeileB = nPaarDavorIndex % ZEILENZAHL;
   nPaarIndex[nPaarIndexAnzahl] = nZeileA<<9 | nZeileB;

   nNaechstesPaarIndex = nPaarDavor[nPaarDavorIndex];
   while(nNaechstesPaarIndex != -1)
   {
      nKey = nNaechstesPaarIndex/nAnzahlPaare;
      nb = nPaarNummer[nKey]++;
      q = nPaarDanach[nNaechstesPaarIndex];

      if(nb < nTestzeilen)
         pKombinationen[nTestzeilen*nKey+nb] = (short)(nPaarIndexAnzahl*2);

      nNaechstesPaarIndex = q;
   }

   nPaarIndexAnzahl++;
}


bool CSyster::SchreibeIndexDatei()
{
   FILE *sFile;

   // Vorberechneten Tabellen in Datei schreiben
   char pDateiName[20];
   
   // Dateiname erzeugen
   sprintf(pDateiName, INDEX_DATEI, nTestzeilen);

   // Datei zum Schreiben ffnen
   sFile = fopen(pDateiName, "wb");

   if(sFile == NULL)
      return false;

   if(fwrite(pKombinationen, sizeof(short), KOMBINATIONEN*nTestzeilen, sFile) != (size_t)(KOMBINATIONEN*nTestzeilen))
   {
      fclose(sFile);
      return false;
   }

   if(fwrite(nZeilenPaarKorrelation, sizeof(int), INDEXSIZE, sFile) != INDEXSIZE)
   {
      fclose(sFile);
      return false;
   }

   if(fwrite(nPaarIndex, sizeof(int), INDEXSIZE, sFile) != INDEXSIZE)
   {
      fclose(sFile);
      return false;
   }

   if(fwrite(&nPaarIndexAnzahl, sizeof(int), 1, sFile) != 1)
   {
      fclose(sFile);
      return false;
   }

   fclose(sFile);
   return true;
}


int CSyster::SucheKey
(
   int   nPalPlusZeile,
   bool  bPalPlus
)
{
   int      nZeilenZlr;
	 int      nKorrelation;
   int      nBesteKorrelation = 0x7fffffff;
   int      nBesterKandidat   = 0;
   short   *pAktuelleKombinationen = pKombinationen;

   // Prfen welche Permutation die Beste ist.
	 //KOMBINATIONEN = 32768
   for(int nPermutationsZlr=0; nPermutationsZlr<KOMBINATIONEN; nPermutationsZlr++)
   {
     //Seb 564654
		 /*
		 if(bPalPlus)
      {
         // Alle Kombinationen berspringen bei denen die erste Zeile nicht die 
         // PalPlusZeile ist. Ist nPalPlusZeile gleich -1, wurde nichts gefunden
         if( nPalPlusZeile != -1 &&
             nPalPlusZeile != pKey[nPermutationsZlr>>7]
           )
         {
            pAktuelleKombinationen+=nTestzeilen;
            continue;
         }
      }*/

      nKorrelation = 0;

			//!!!
      //for(int nZeile=0; nZeile<nTestzeilen; nZeile++)
			for(nZeilenZlr=nTestzeilen-1; nZeilenZlr>=0;nZeilenZlr--)
			{
         nKorrelation += *((unsigned short*)nZeilenPaarKorrelation+*pAktuelleKombinationen++);

         /*if(nKorrelation > nBesteKorrelation)
         {
            // Zeiger korrigieren und die nchste Permutation testen
            pAktuelleKombinationen+=nZeilenZlr;
            break;
         }*/
			}
			//!!!

      if(nKorrelation < nBesteKorrelation)
      {
         nBesteKorrelation = nKorrelation;
         nBesterKandidat = nPermutationsZlr;
      }
   }

	 /*_asm {
     pusha

     mov ecx, 32768
		 //mov ecx, 8192
		 xor edx, edx // nPermutationsZlr

     //Korrelation = 0;
     mov esi, dword ptr pKombinationen
		 lea edi, nZeilenPaarKorrelation

   BoucleSuchKey:
		 xor ebx, ebx // Reprsente nKorrelation
		 push ecx
     
 		  //!!!
      //for(int nZeile=0; nZeile<nTestzeilen; nZeile++)
      //   nKorrelation += *((unsigned short*)nZeilenPaarKorrelation+*pAktuelleKombinationen++);
			//!!!

		 mov ecx, nTestzeilen
		 //mov ecx, 32

   BoucleKorrelation:
		 xor eax, eax
		 lodsw

		 mov ax, word ptr [edi+eax]
		 add ebx, eax
		 //add bx, word ptr [edi+eax]
		 loop BoucleKorrelation

     //if(nKorrelation < nBesteKorrelation)
     //{
     //  nBesteKorrelation = nKorrelation;
     //  nBesterKandidat = nPermutationsZlr;
     //}

     //mov ebx, nKorrelation
		 //mov ecx, nBesteKorrelation
		 cmp ebx, nBesteKorrelation
		 jge Fin

		 mov nBesteKorrelation, ebx
		 //mov ebx, nPermutationsZlr
		 mov nBesterKandidat, edx

   Fin:
		 pop ecx
		 inc edx
		 loop BoucleSuchKey

		 popa
	 }*/

   return nBesterKandidat;
}


CYVUSyster::CYVUSyster(int nTestzeilen) : CSyster(nTestzeilen)
{
}


CYVUSyster::~CYVUSyster()
{
}


void CYVUSyster::DecodeColor
(
   int **pLinesAddress,
   int  *pSortLines,
   int  *pDestination,
   int   nSourceWidth,
   int   nDestinationWidth,
   int   nDestinationHeight,
   int   nPremiereLigne,
   int   nDerniereLigne,
   bool  bPalPlus
)
{
   int            nPhase;
   int            nPixel;
   int            nZeileZiel;
   int            nZeileQuelle;
   int            nLastPhase     = 4;
   int            nLastColorLine = 0;
   unsigned int  *pZiel;
   unsigned int  *pQuelle;
   unsigned int  *pColorQuelle;

   //
   // Die Daten kommen im YUY2 Format:
   //
   //  31--V0--24 23--Y1--16 15--U0--8 7--Y0--0
   //  MSB                                  LSB
   //
   // U0 und V0 gelten jeweils fr Y0 und Y1
   //

   // Zeilen sortieren
   for(nZeileZiel=nPremiereLigne; nZeileZiel<nDerniereLigne; nZeileZiel++)
   {
      nZeileQuelle = pSortLines[nZeileZiel] - 32;
      pZiel        = (unsigned int*)&pDestination[nZeileZiel*nDestinationWidth/4];
      pQuelle      = (unsigned int*)pLinesAddress[pSortLines[nZeileZiel]];
      pColorQuelle = (unsigned int*)pLinesAddress[pSortLines[nLastColorLine]];

      if(nZeileQuelle > 0)
      {
         nPhase = (nZeileZiel-nZeileQuelle)&0x03;
      }
      else
      {
         if(nZeileQuelle == 0)
         {
            // Die fehlende Zeile ohne Farbe anzeigen
            nPhase = 4;
         }
         else
         {  
            // Bei den ersten 32 Zeilen Farbe invertieren. Bei PALplus kommt das 
            // erste Halbbild (+1), sonst das zweite (+2).
            if(bPalPlus)
               nPhase = (nZeileZiel+1-nZeileQuelle)&0x03;
            else
               nPhase = (nZeileZiel+2-nZeileQuelle)&0x03;
         }
      }

      // Wenn der Abstand zu gro wird die Farbe aus der letzten gltigen Zeile holen
      if(nZeileZiel-nZeileQuelle > 40)
         nPhase = 4;

      // Farbe korrigieren
      switch(nPhase)
      {
         case 0:
            nLastPhase     = nPhase;
            nLastColorLine = nZeileZiel;

            // v = v, u = u
            for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
            {
               *pZiel = *pQuelle;

               pZiel++;
               pQuelle++;
            }
         break;

         case 2:
            nLastPhase     = nPhase;
            nLastColorLine = nZeileZiel;

            // v = -v, u = -u
            for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
            {
               *pZiel = *pQuelle^0xff00ff00;

               pZiel++;
               pQuelle++;
            }
         break;

         default:
            switch(nLastPhase)
            {
               case 0:
                  // v = v, u = u
                  for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
                  {
                     *pZiel = (*pQuelle&0x00ff00ff) | (*pColorQuelle&0xff00ff00);

                     pZiel++;
                     pQuelle++;
                     pColorQuelle++;
                  }
               break;

               case 2:
                  // v = -v, u = -u
                  for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
                  {
                     *pZiel = (*pQuelle&0x00ff00ff) | ((*pColorQuelle&0xff00ff00)^0xff00ff00);

                     pZiel++;
                     pQuelle++;
                     pColorQuelle++;
                  }
               break;

               default:
                  // Alle andern Zeilen ohne Farbe anzeigen
                  for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
                  {
                     *pZiel = (*pQuelle&0x00ff00ff) | 0x80008000;

                     pZiel++;
                     pQuelle++;
                  }
            }
         break;
      }
   }
}


/*
 * Das PAL (Phase-Alternate-Line) System ist eine Weiterentwicklung des
 * NTSC-Verfahrens. Dessen Nachteile (Farbtonfehler) werden mit elektronischen 
 * Mehraufwand gut kompensiert.
 * 
 * Encoder
 * 
 * Das Leuchtdichtesignal (Y) wird aus den Farbwertsignalen R,B und G (rot, grn
 * und blau) unter Bercksichtigung der Empfindlichkeit des Auges fr die
 * einzelnen Farben gebildet:
 * 
 * Y = 0.3*R+0.59*G+0.11*B
 * 
 * Aus Kompatibilittsgrnden werden bei allen gngigen Farbbertragungsverfahren
 * nur die Farbdifferenzsignale bertragen:
 * 
 * U = 0.493 (B-Y)
 * V = 0.877 (R-Y)
 * 
 * Diese Differenzsignale werden beim PAL-System zur bertragung auf den 
 * Farbtrger (Fc) moduliert: 
 * 
 * Uc = U * sin(Fc*2*Pi)
 * Vc = V * cos(Fc*2*Pi)
 * 
 * Zustzlich invertiert man Vc jede zweite Zeile. Die Referenzphase
 * des Farbtrgers wird auf der hinteren Schwarzschulter des
 * Zeilensynchronimpulses fr 2.26 s bertragen. Die Phasenlage
 * diese Burst ist: 
 * 
 * 135 fr Zeilen mit +Vc
 * 225 fr Zeilen mit -Vc
 * 
 * Die Farbtrgerfrequenz wurde mit 4.4336175 MHz festgelegt, d.h. eine Zeile (64s) 
 * entspricht 283.75 Farbtrgerschwingungen. 
 * 
 * Alle Signale (Y, Uc, Vc) werden addiert und stellen das Farbartsignal dar.
 * 
 * Decoder
 * 
 * Im Decoder wird das Y-Signal durch einen Bandsperre von Uc und Vc befreit.
 * Umgekehrt erhlt man das Croma-Signal C = Uc+Vc hinter einem Bandpass. 
 * 
 * Zur Rckgewinnung der Referenzphasenlage des Farbtrgers wird der Burst ber
 * eine Torschaltung einer Synchronisationsstufe zugeleitet. Sie stellt einen
 * Phasendiskriminator dar, der zwei Ausgnge hat: der Ausgang mit hoher
 * Zeitkonstante syncronisiert den Referenzoszillator, aus dem Ausgang mit 
 * niedriger Zeitkonstante erhlt man die Phasenlage von Vc.
 * 
 * Die eigendlich Demodulation und damit die Rckgewinnung von U und V aus
 * Vc und Uc erfolgt in einem sogenannten Laufzeitdemodulator. C wird mit einer
 * Verzgerungsleitung um 63,943 s (180 Phasendrehung) verzgert, so da man 
 * die aktuelle Zeile Cn und die vorherige Zeile Cn-1 phsengedreht gleichzeitig 
 * zur Verfgung hat. 
 * 
 * Durch die Addition bzw. Subtrakion von Cn und Cn-1 erhlt man:
 * 
 * 2 Uc' = Cn + Cn-1 = (Uc+Vc)+((Uc-1)-(Vc-1))
 * 
 * und je nach Phasenlage von Vc in der aktuellen Zeile
 * 
 * 2 Vc' = Cn - Cn-1 = (Uc+Vc)-((Uc-1)-(Vc-1))
 * 
 * oder
 * 
 * -2 Vc' = Cn - Cn-1 = (Uc-Vc)-((Uc-1)+(Vc-1))
 * 
 * Uc' und Vc' werden also aus zwei benachbarten Zeilen gemittelt. Wie man
 * leicht nachweisen kann, wird aber eine Phasenverschiebung von C durch dieses
 * Verfahren in einen Farbsttigungsfehler umgewandelt. Dieser ist weniger
 * strend als die von einem Phasenfehler hervorgerufenen Farbtonfehler.
 * 
 * Nach der Multiplikation mit der 0 Phase des Referenzfarbtrgers erhlt man U.
 * Die V-Komponente ergibt sich durch die Multiplikation mit der 90 bzw. 180
 * Phase in Abhngigkeit von der Phasenlage von Vc. 
 * 
 * Um die Farbwertsignale zu erhalten, ist noch folgende Transformation ntig:
 * 
 * R = Y + V
 * B = Y + U
 * G = 1,7*Y-0,51*R-0,19*B 
 * 
 * Fehler bei der PAL-Decodierung eines verschlsselten Syster-Signals
 * 
 * Das Systerverfahren beruht auf der scheinbar willkrlichen Verwrfelung
 * von kompletten Bildzeilen. Wird das verschlsselte Bild mit einer
 * Framegrabberkarte aufgezeichnet, erfolgt die PAL-Decodierung vor der 
 * Entschlsselung. Es ergeben sich im wesentlichen zwei Fehler:
 * 
 * 1. Bei der Rckgewinnung der U und V-Komponenten im Laufzeitdemodulator
 * werden die Farbkomponenten von nicht zusammengehrige Zeilen addiert 
 * bzw. subtrahiert. Es entstehen Fehler in Betrag und Phase, die unter
 * Umstnden sogar zur Auslschung der Farbe fhren knnen. Sie sind daher
 * auch durch nachtgliche rechnerische Bearbeitung der Farbkomponenten nicht
 * mehr zu korrigieren.
 * 
 * Wenn die Farb-Decodierung vor der Entschlsselung des Bildes erfolgt, kann
 * man dieses Problem nur durch Abschaltung des Laufzeitdemodulators umgehen.
 * Nun werden allerdings eventuelle Farbtonfehler nicht mehr automatisch
 * korrigiert. Innerhalb der Framegrabber-IC BT848 kann der 'digitale 
 * Laufzeitdemodulator' abgeschalten werden.
 * 
 * Man kann natrlich auch die Farbdecodierung durch Software nach der
 * Syster-Decodierung machen, da wird dann aber seeeeeehr langsam.
 * 
 * 2. Da innerhalb einer Zeile 283.75 Farbtrgerschwingungen auftreten, ist die 
 * Phase jeder Zeile gegenber der vorhergenden um 90 verschoben. Das fhrt
 * zu massiven Farbtonfehlern. Daher stammen auch die typischen farbigen Streifen
 * im verschlsselten Syster-Bild.
 * 
 * Der Burst wird im verschlsselten Bild nicht getauscht (warum wohl ;-), 
 * so da die Referenzphase fehlerfrei zurckgewonnen werden kann. Wenn man mit
 * einer Framegrabberkarte die Bilder im YUV-Format abspeichert (je zwei Pixel
 * haben die selben Farbe U und V) kann man bei Kenntnis der Verschiebung
 * jeder einzelnen Zeile zwischen verschlsselten und unverschlsselten Bild
 * die ursprngliche Phase rekonstruieren. Da die Phase der Farbkomponenten 
 * immer um Vielfache von 90 gedreht ist, reduziert sich das Ganze auf richtiges 
 * Vertauschung oder/und den Vorzeichenwechsel der U und V-Bytes. Wird nun noch 
 * die YUV nach RGB-Konvertierung durch Software gemacht, braucht sich ein P200 
 * nicht zu langweilen.
 */

CRGBSyster::CRGBSyster
(
   int nTestzeilen, 
   int nPixelMode
) : CSyster(nTestzeilen)
{
   m_pYUV2RGBTabelle = new short[YUV2RGB_SIZE];
   if(m_pYUV2RGBTabelle == NULL)
      return;

   // Tabellen zur Farbkorrektur erstellen
   BaueYUV2RGBTabelle(nPixelMode);
}


CRGBSyster::~CRGBSyster()
{
   if(m_pYUV2RGBTabelle != NULL)
      delete[] m_pYUV2RGBTabelle;
}


void CRGBSyster::BaueYUV2RGBTabelle
(
   int nPixelMode
)
{
   int nY;
   int nV;
   int nU;
   int nR;
   int nG;
   int nB;

   if(m_pYUV2RGBTabelle == NULL)
      return;

   //
   // Das Format der Tabelle YUV2RGB:
   //
   //   0x3F00    0x0FC0  0x003F
   //  17--V--12 11--U--6 5--Y--0
   //  MSB                    LSB
   //
	for(nY=0; nY<256; nY+=4)  
	{
		for(nU=0; nU<256; nU+=4)
    {
			for(nV=0; nV<256; nV+=4)
      {
        nB = (116*(nY-16)                + 202*(nU-128)) / 100;
        nG = (116*(nY-16) -  81*(nV-128) -  39*(nU-128)) / 100;
        nR = (116*(nY-16) + 160*(nV-128)               ) / 100;

        if(nG>0xff) nG=0xff;

        if(nB>0xff) nB=0xff;

        if(nR>0xff) nR=0xff;

        if(nG<0) nG=0;

        if(nB<0) nB=0;

        if(nR<0) nR=0;

        if(nPixelMode == GRAFIKMODE_RGB15)
          m_pYUV2RGBTabelle[((nV>>2)<<12) | ((nU>>2)<<6) | nY>>2] = (short)((nB>>3) | ((nG>>3)<<5) | ((nR>>3)<<10));

				if(nPixelMode == GRAFIKMODE_RGB16)
          m_pYUV2RGBTabelle[((nV>>2)<<12) | ((nU>>2)<<6) | nY>>2] = (short)((nB>>3) | ((nG>>2)<<5) | ((nR>>3)<<11));
 
        //if(nPixelMode == GRAFIKMODE_RGB24)
        //  m_pYUV2RGBTabelle[((nV>>2)<<12) | ((nU>>2)<<6) | nY>>2] = nB | (nG<<8) | (nR<<16);
			}      
      }
   }
}


void CRGBSyster::DecodeColor
(
   int **pLinesAddress,
   int  *pSortLines,
   int  *pDestination,
   int   nSourceWidth,
   int   nDestinationWidth,
   int   nDestinationHeight,
   int   nPremiereLigne,
   int   nDerniereLigne,
   bool  bPalPlus
)
{
   int            nPixel;
   int            nPhase;
   int            nZeileZiel;
   int            nZeileQuelle;
   int            nScreenLine;
   int            nLastPhase     = 4;
   int            nLastColorLine = 0;
   unsigned int   nUV;
   unsigned int   nRGB1;
   unsigned int   nRGB2;
   unsigned int  *pZiel;
   unsigned int  *pQuelle;
   unsigned int  *pColorQuelle;

   //
   // Die Daten kommen im YUY2 Format:
   //
   //  0xFF000000 0x00FF0000 0x000FF00 0x0000FF
   //  31--V0--24 23--Y1--16 15--U0--8 7--Y0--0
   //  MSB                                  LSB
   //
   // U0 und V0 gelten jeweils fr Y0 und Y1
   //

   if(m_pYUV2RGBTabelle == NULL)
      return;

   // Zeilen sortieren
   nScreenLine = nPremiereLigne*nDestinationHeight/ZEILENZAHL;
   for(nZeileZiel=nPremiereLigne; nZeileZiel<nDerniereLigne; nZeileZiel++)
   {
      nZeileQuelle = pSortLines[nZeileZiel] - 32;

      if(nZeileQuelle > 0)
      {
         nPhase = (nZeileZiel-nZeileQuelle)&0x03;
      }
      else
      {
         if(nZeileQuelle == 0)
         {
            // Die fehlende Zeile ohne Farbe anzeigen
            nPhase = 4;
         }
         else
         {  
            // Bei den ersten 32 Zeilen Farbe invertieren. Bei PALplus kommt das 
            // erste Halbbild (+1), sonst das zweite (+2).
            if(bPalPlus)
               nPhase = (nZeileZiel+1-nZeileQuelle)&0x03;
            else
               nPhase = (nZeileZiel+2-nZeileQuelle)&0x03;
         }
      }

      // Wenn der Abstand zu gro wird die Farbe aus der letzten gltigen Zeile holen
      if(nZeileZiel-nZeileQuelle > 40)
         nPhase = 4;

      // Die Bild auf Bildschirmhhe strecken
      while(nScreenLine < nZeileZiel*nDestinationHeight/ZEILENZAHL)
      {
#ifdef DIRECTX
         pZiel = (unsigned int*)(pDestination+nScreenLine*nDestinationWidth/4);
#else
         pZiel = (unsigned int*)(pDestination+(nDestinationHeight-1-nScreenLine)*nDestinationWidth/4);
#endif
         pQuelle      = (unsigned int*)pLinesAddress[pSortLines[nZeileZiel]];
         pColorQuelle = (unsigned int*)pLinesAddress[pSortLines[nLastColorLine]];

         // Farbe korrigieren
         switch(nPhase)
         {
            case 0:
               nLastPhase     = nPhase;
               nLastColorLine = nZeileZiel;

               // v = v, u = u
               for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
               {
                  nUV = *pQuelle>>4 & 0xfc0 | *pQuelle>>14 & 0x3f000;

		            nRGB1 = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | nUV];
                  nRGB2 = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | nUV];

                  *pZiel++ = nRGB1 | nRGB2<<16;
                  pQuelle++;
               }
            break;

            case 2:
               nLastPhase     = nPhase;
               nLastColorLine = nZeileZiel;

               // v = -v, u = -u
               for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
               {
                  nUV = *pQuelle^0xff00ff00;
                  nUV = nUV>>4 & 0xfc0 | nUV>>14 & 0x3f000;

		            nRGB1 = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | nUV];
                  nRGB2 = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | nUV];

                  *pZiel++ = nRGB1 | nRGB2<<16;
                  pQuelle++;
               }
            break;

            default:
               switch(nLastPhase)
               {
                  case 0:
                     // v = v, u = u
                     for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
                     {
                        nUV = *pColorQuelle>>4 & 0xfc0 | *pColorQuelle>>14 & 0x3f000;

		                  nRGB1 = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | nUV];
                        nRGB2 = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | nUV];

                        *pZiel++ = nRGB1 | nRGB2<<16;
                        pQuelle++;
                        pColorQuelle++;
                     }
                  break;

                  case 2:
                     // v = -v, u = -u
                     for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
                     {
                        nUV = *pColorQuelle^0xff00ff00;
                        nUV = nUV>>4 & 0xfc0 | nUV>>14 & 0x3f000;

		                  nRGB1 = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | nUV];
                        nRGB2 = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | nUV];

                        *pZiel++ = nRGB1 | nRGB2<<16;
                        pQuelle++;
                        pColorQuelle++;
                     }
                  break;

                  default:
                     // Alle andern Zeilen ohne Farbe anzeigen
                     for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
                     {
		                nRGB1 = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | 0x20800];
                        nRGB2 = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | 0x20800];
                        *pZiel++ = nRGB1 | nRGB2<<16;
                        pQuelle++;
                     }
               }
            break;
         }

         nScreenLine++;
      }
   }
}


//
// Secam Decodierung von Xetra
//
CYVUSysterSecam::CYVUSysterSecam(int nTestzeilen) : CSyster(nTestzeilen)
{
}


CYVUSysterSecam::~CYVUSysterSecam()
{
}


void CYVUSysterSecam::DecodeColor
(
   int **pLinesAddress,
   int  *pSortLines,
   int  *pDestination,
   int   nSourceWidth,
   int   nDestinationWidth,
   int   nDestinationHeight,
   int   nPremiereLigne,
   int   nDerniereLigne,
   bool  bPalPlus
)
{
   int            nPixel;
   int            nZeileZiel;
   unsigned int  *pZiel;
   unsigned int  *pQuelle;
   
   // By Xetra
   /*pZiel   = (unsigned int*)&pDestination[nPremiereLigne*nDestinationWidth/4];
    for(nZeileZiel=nPremiereLigne; nZeileZiel<nDerniereLigne; nZeileZiel++)
   {
 	  pQuelle = (unsigned int*)pLinesAddress[pSortLines[nZeileZiel]];      
 	  CopyMemory( pZiel, pQuelle, nSourceWidth<<2 );
	  pZiel+=nSourceWidth>>1;
   }*/

   
   // Zeilen sortieren
   for(nZeileZiel=nPremiereLigne; nZeileZiel<nDerniereLigne; nZeileZiel++)
   {
      pZiel   = (unsigned int*)&pDestination[nZeileZiel*nDestinationWidth/4];
      pQuelle = (unsigned int*)pLinesAddress[pSortLines[nZeileZiel]];
      // v = v, u = u
      for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
	  { 
		 *pZiel = *pQuelle;
         pZiel++;
         pQuelle++;
	  }	  
	  //*pZiel++ = *pQuelle++;//Xetra
   }
}


//
// Secam Decodierung von Xetra
//
CRGBSysterSecam::CRGBSysterSecam
(
   int nTestzeilen, 
   int nPixelMode
) : CSyster(nTestzeilen)
{
   m_pYUV2RGBTabelle = new short [YUV2RGB_SIZE];
   if(m_pYUV2RGBTabelle == NULL)
      return;

   // Tabellen zur Farbkorrektur erstellen
   BaueYUV2RGBTabelle(nPixelMode);
}


CRGBSysterSecam::~CRGBSysterSecam()
{
   if(m_pYUV2RGBTabelle != NULL)
      delete[] m_pYUV2RGBTabelle;
}


void CRGBSysterSecam::BaueYUV2RGBTabelle
(
   int nPixelMode
)
{
   int nY;
   int nV;
   int nU;
   int nR;
   int nG;
   int nB;
   
   if(m_pYUV2RGBTabelle == NULL)
      return;

   //
   // Das Format der Tabelle YUV2RGB:
   //
   //   0x3F00    0x0FC0  0x003F
   //  17--V--12 11--U--6 5--Y--0
   //  MSB                    LSB
   //

	 // Modif Seb 6/11/1998
 	 for(nV=0; nV<256; nV++)
   {
		 for(nU=0; nU<256; nU++)
     {
	     for(nY=0; nY<256; nY++)
       {
            nB = (116*(nY-16)                + 202*(nU-128)) / 100;
            nG = (116*(nY-16) -  81*(nV-128) -  39*(nU-128)) / 100;
            nR = (116*(nY-16) + 160*(nV-128)               ) / 100;
            //nB = (116*(nY-16))/ 100;
            //nG = (116*(nY-16))/ 100;
            //nR = (116*(nY-16))/ 100;
			           
			      if(nG>0xff) nG=0xff;
            if(nB>0xff) nB=0xff;
            if(nR>0xff) nR=0xff;
            if(nG<0) nG=0;
            if(nB<0) nB=0;
            if(nR<0) nR=0;
            
/*            if(nPixelMode == GRAFIKMODE_RGB15)
              m_pYUV2RGBTabelle[((nV>>2)<<12) | ((nU>>2)<<6) | nY>>2] = 
							  (short)((nB>>3) | ((nG>>3)<<5) | ((nR>>3)<<10));*/

//            if(nPixelMode == GRAFIKMODE_RGB16)
              /*m_pYUV2RGBTabelle[((nY>>6)<<10) | ((nU>>6)<<8) | nY] = 
							  (short)((nB>>3) | ((nG>>2)<<5) | ((nR>>3)<<11));*/
								YU12toRGB[nY | (nV<<8) | (nU<<16)] =
							  (short)((nB>>3) | ((nG>>3)<<5) | ((nR>>3)<<10));

            /*if(nPixelMode == GRAFIKMODE_RGB24)
              m_pYUV2RGBTabelle[((nV>>2)<<12) | ((nU>>2)<<6) | nY>>2] = 
							  (short)((nB>>3) | ((nG>>3)<<5) | ((nR>>3)<<10));*/

			}
      }
   }

	// Modif Seb 6/11/1998
}


/*   pSyster->DecodeColor( pLinesAddress,
                         pSortLines,
                         (int*)sDirectDrawSurfaceDesc.lpSurface,
								 pStartDialog->VideoWidth(),
                         nStartOfNextLine,
                         pStartDialog->ZeilenAusgabe(),
                         pStartDialog->GetErsteZeile(),
                         pStartDialog->GetLetzteZeile(),
                         pStartDialog->PalPlus()
                       );*/

void CRGBSysterSecam::DecodeColor
(
   int **pLinesAddress,
   int  *pSortLines,
   int  *pDestination, //pDestination
   int   nSourceWidth, //352
   int   nDestinationWidth, //400
   int   nDestinationHeight, //300
   int   nPremiereLigne, // Premiere Ligne
   int   nDerniereLigne, // Derniere Ligne
   bool  bPalPlus  //FALSE
)
{
   int            nPixel;
   int            nZeileZiel;
   int            nScreenLine;
   unsigned int   nUV;
   unsigned int   nRGB1;
   unsigned int   nRGB2;
   unsigned int  *pQuelle;
   unsigned int  *pZiel;

   if(m_pYUV2RGBTabelle == NULL)
      return;

   // Zeilen sortieren
   nScreenLine = nPremiereLigne*nDestinationHeight/ZEILENZAHL;

   for(nZeileZiel=nPremiereLigne; nZeileZiel<nDerniereLigne; nZeileZiel++)
   {
      // Die Bild auf Bildschirmhhe strecken
      while(nScreenLine < nZeileZiel*nDestinationHeight/ZEILENZAHL)
      {
#ifdef DIRECTX
         pZiel = (unsigned int*)(pDestination+nScreenLine*nDestinationWidth/4);
#else
         pZiel = (unsigned int*)(pDestination+(nDestinationHeight-1-nScreenLine)*nDestinationWidth/4);
#endif
         pQuelle = (unsigned int*)pLinesAddress[pSortLines[nZeileZiel]];

         for(nPixel=0; nPixel<nSourceWidth; nPixel+=2)
         {
            //nUV      = *pQuelle>>4 & 0xfc0 | *pQuelle>>14 & 0x3f000;
		    //nRGB1    = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | nUV];
          //  nRGB2    = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | nUV];
            //*pZiel++ = nRGB1 | nRGB2<<16;
					 *pZiel++ = *pQuelle;
            pQuelle++;
         }

         nScreenLine++;
      }
   }
}
