#pragma warning(disable: 4056 4100 4201 4514)

#include <windows.h>

#include "RGBSyster.h"
#include "Debug.h"
#include "resource.h"


/*
 * Das PAL (Phase-Alternate-Line) System ist eine Weiterentwicklung des
 * NTSC-Verfahrens. Dessen Nachteile (Farbtonfehler) werden mit elektronischen 
 * Mehraufwand gut kompensiert.
 * 
 * Encoder
 * 
 * Das Leuchtdichtesignal (Y) wird aus den Farbwertsignalen R,B und G (rot, grn
 * und blau) unter Bercksichtigung der Empfindlichkeit des Auges fr die
 * einzelnen Farben gebildet:
 * 
 * Y = 0.3*R+0.59*G+0.11*B
 * 
 * Aus Kompatibilittsgrnden werden bei allen gngigen Farbbertragungsverfahren
 * nur die Farbdifferenzsignale bertragen:
 * 
 * U = 0.493 (B-Y)
 * V = 0.877 (R-Y)
 * 
 * Diese Differenzsignale werden beim PAL-System zur bertragung auf den 
 * Farbtrger (Fc) moduliert: 
 * 
 * Uc = U * sin(Fc*2*Pi)
 * Vc = V * cos(Fc*2*Pi)
 * 
 * Zustzlich invertiert man Vc jede zweite Zeile. Die Referenzphase
 * des Farbtrgers wird auf der hinteren Schwarzschulter des
 * Zeilensynchronimpulses fr 2.26 s bertragen. Die Phasenlage
 * diese Burst ist: 
 * 
 * 135 fr Zeilen mit +Vc
 * 225 fr Zeilen mit -Vc
 * 
 * Die Farbtrgerfrequenz wurde mit 4.4336175 MHz festgelegt, d.h. eine Zeile (64s) 
 * entspricht 283.75 Farbtrgerschwingungen. 
 * 
 * Alle Signale (Y, Uc, Vc) werden addiert und stellen das Farbartsignal dar.
 * 
 * Decoder
 * 
 * Im Decoder wird das Y-Signal durch einen Bandsperre von Uc und Vc befreit.
 * Umgekehrt erhlt man das Croma-Signal C = Uc+Vc hinter einem Bandpass. 
 * 
 * Zur Rckgewinnung der Referenzphasenlage des Farbtrgers wird der Burst ber
 * eine Torschaltung einer Synchronisationsstufe zugeleitet. Sie stellt einen
 * Phasendiskriminator dar, der zwei Ausgnge hat: der Ausgang mit hoher
 * Zeitkonstante syncronisiert den Referenzoszillator, aus dem Ausgang mit 
 * niedriger Zeitkonstante erhlt man die Phasenlage von Vc.
 * 
 * Die eigendlich Demodulation und damit die Rckgewinnung von U und V aus
 * Vc und Uc erfolgt in einem sogenannten Laufzeitdemodulator. C wird mit einer
 * Verzgerungsleitung um 63,943 s (180 Phasendrehung) verzgert, so da man 
 * die aktuelle Zeile Cn und die vorherige Zeile Cn-1 phsengedreht gleichzeitig 
 * zur Verfgung hat. 
 * 
 * Durch die Addition bzw. Subtrakion von Cn und Cn-1 erhlt man:
 * 
 * 2 Uc' = Cn + Cn-1 = (Uc+Vc)+((Uc-1)-(Vc-1))
 * 
 * und je nach Phasenlage von Vc in der aktuellen Zeile
 * 
 * 2 Vc' = Cn - Cn-1 = (Uc+Vc)-((Uc-1)-(Vc-1))
 * 
 * oder
 * 
 * -2 Vc' = Cn - Cn-1 = (Uc-Vc)-((Uc-1)+(Vc-1))
 * 
 * Uc' und Vc' werden also aus zwei benachbarten Zeilen gemittelt. Wie man
 * leicht nachweisen kann, wird aber eine Phasenverschiebung von C durch dieses
 * Verfahren in einen Farbsttigungsfehler umgewandelt. Dieser ist weniger
 * strend als die von einem Phasenfehler hervorgerufenen Farbtonfehler.
 * 
 * Nach der Multiplikation mit der 0 Phase des Referenzfarbtrgers erhlt man U.
 * Die V-Komponente ergibt sich durch die Multiplikation mit der 90 bzw. 180
 * Phase in Abhngigkeit von der Phasenlage von Vc. 
 * 
 * Um die Farbwertsignale zu erhalten, ist noch folgende Transformation ntig:
 * 
 * R = Y + V
 * B = Y + U
 * G = 1,7*Y-0,51*R-0,19*B 
 * 
 * Fehler bei der PAL-Decodierung eines verschlsselten Syster-Signals
 * 
 * Das Systerverfahren beruht auf der scheinbar willkrlichen Verwrfelung
 * von kompletten Bildzeilen. Wird das verschlsselte Bild mit einer
 * Framegrabberkarte aufgezeichnet, erfolgt die PAL-Decodierung vor der 
 * Entschlsselung. Es ergeben sich im wesentlichen zwei Fehler:
 * 
 * 1. Bei der Rckgewinnung der U und V-Komponenten im Laufzeitdemodulator
 * werden die Farbkomponenten von nicht zusammengehrige Zeilen addiert 
 * bzw. subtrahiert. Es entstehen Fehler in Betrag und Phase, die unter
 * Umstnden sogar zur Auslschung der Farbe fhren knnen. Sie sind daher
 * auch durch nachtgliche rechnerische Bearbeitung der Farbkomponenten nicht
 * mehr zu korrigieren.
 * 
 * Wenn die Farb-Decodierung vor der Entschlsselung des Bildes erfolgt, kann
 * man dieses Problem nur durch Abschaltung des Laufzeitdemodulators umgehen.
 * Nun werden allerdings eventuelle Farbtonfehler nicht mehr automatisch
 * korrigiert. Innerhalb der Framegrabber-IC BT848 kann der 'digitale 
 * Laufzeitdemodulator' abgeschalten werden.
 * 
 * Man kann natrlich auch die Farbdecodierung durch Software nach der
 * Syster-Decodierung machen, da wird dann aber seeeeeehr langsam.
 * 
 * 2. Da innerhalb einer Zeile 283.75 Farbtrgerschwingungen auftreten, ist die 
 * Phase jeder Zeile gegenber der vorhergenden um 90 verschoben. Das fhrt
 * zu massiven Farbtonfehlern. Daher stammen auch die typischen farbigen Streifen
 * im verschlsselten Syster-Bild.
 * 
 * Der Burst wird im verschlsselten Bild nicht getauscht (warum wohl ;-), 
 * so da die Referenzphase fehlerfrei zurckgewonnen werden kann. Wenn man mit
 * einer Framegrabberkarte die Bilder im YUV-Format abspeichert (je zwei Pixel
 * haben die selben Farbe U und V) kann man bei Kenntnis der Verschiebung
 * jeder einzelnen Zeile zwischen verschlsselten und unverschlsselten Bild
 * die ursprngliche Phase rekonstruieren. Da die Phase der Farbkomponenten 
 * immer um Vielfache von 90 gedreht ist, reduziert sich das Ganze auf richtiges 
 * Vertauschung oder/und den Vorzeichenwechsel der U und V-Bytes. Wird nun noch 
 * die YUV nach RGB-Konvertierung durch Software gemacht, braucht sich ein P200 
 * nicht zu langweilen.
 */

CRGBSyster::CRGBSyster
(
   int   nBreiteQuelle,
   int   nHoeheZiel,
   int   nTestzeilen,
   bool  bRGB15,
   bool  bPalPlus,
   bool  bSchwarzeZeile,
   bool  bSecam
) : CSyster(nTestzeilen, bSchwarzeZeile)
{
   m_nBreiteQuelle   = nBreiteQuelle;
   m_bPALplus        = bPalPlus;

   m_nHoeheZiel      = nHoeheZiel;
   m_bSecam          = bSecam;
	m_pYUV2RGBTabelle = NULL;

   m_pYUV2RGBTabelle = new USHORT[64*64*64];
   if(m_pYUV2RGBTabelle == NULL)
   {
      MessageBox(IDS_ERR_MEMORY);
      m_bError = true;
      return;
   }

   // Tabellen zur Farbkorrektur erstellen
   BaueYUV2RGBTabelle(bRGB15);
}


CRGBSyster::~CRGBSyster()
{
   if(m_pYUV2RGBTabelle != NULL)
      delete[] m_pYUV2RGBTabelle;
}


void CRGBSyster::BaueYUV2RGBTabelle
(
   bool  bRGB15
)
{
   int nY;
   int nV;
   int nU;
   int nR;
   int nG;
   int nB;

   //
   // Das Format der Tabelle YUV2RGB:
   //
   //   0x3F00    0x0FC0  0x003F
   //  17--V--12 11--U--6 5--Y--0
   //  MSB                    LSB
   //
   for(nY=0; nY<256; nY+=4)  
   {
      for(nU=0; nU<256; nU+=4)
      {
         for(nV=0; nV<256; nV+=4)
         {
            nB = (116*(nY-16)                + 202*(nU-128)) / 100;
            nG = (116*(nY-16) -  81*(nV-128) -  39*(nU-128)) / 100;
            nR = (116*(nY-16) + 160*(nV-128)               ) / 100;

            if(nG>0xff)
               nG=0xff;

            if(nB>0xff)
               nB=0xff;

            if(nR>0xff)
               nR=0xff;

            if(nG<0)
               nG=0;

            if(nB<0)
               nB=0;

            if(nR<0)
               nR=0;

            if(bRGB15)
               m_pYUV2RGBTabelle[((nV>>2)<<12) | ((nU>>2)<<6) | nY>>2] = (USHORT)((nB>>3) | ((nG>>3)<<5) | ((nR>>3)<<10));
            else
               m_pYUV2RGBTabelle[((nV>>2)<<12) | ((nU>>2)<<6) | nY>>2] = (USHORT)((nB>>3) | ((nG>>2)<<5) | ((nR>>3)<<11));
         }
      }
   }
}


void CRGBSyster::DecodeColor
(
   int **pLinesAddress,
   int  *pSortLines,
   int  *pDestination,
   int   nBreiteZiel,
   int   nErsteSichtbareZeile,
   int   nLetzteSichtbareZeile,
   bool  bBild,
   bool  bDecode,
   bool  bPhase
)
{
   int    nPixel;
   int    nPhase;
   int    nZeileZiel;
   int    nZeileQuelle;
   int    nScreenLine;
   int    nLastPhase     = 4;
   int    nLastColorLine = 0;
   DWORD  nUV;
   DWORD  nRGB1;
   DWORD  nRGB2;
   DWORD *pZiel;
   DWORD *pQuelle;
   DWORD *pColorQuelle;

   //
   // Die Daten kommen im YUY2 Format:
   //
   //  0xFF000000 0x00FF0000 0x000FF00 0x0000FF
   //  31--V0--24 23--Y1--16 15--U0--8 7--Y0--0
   //  MSB                                  LSB
   //
   // U0 und V0 gelten jeweils fr Y0 und Y1
   //

   // Zeilen sortieren
   nScreenLine = nErsteSichtbareZeile*m_nHoeheZiel/ZEILENZAHL;
   for(nZeileZiel=nErsteSichtbareZeile; nZeileZiel<nLetzteSichtbareZeile; nZeileZiel++)
   {
      if(bDecode == false)
      {
         nPhase = 0;
      }
      else
      {
         //
         // Secam Decodierung von Xetra
         //
         if(m_bSecam)
         {
            nPhase = 0;
         }
         else
         {
            nZeileQuelle = pSortLines[nZeileZiel] - 32;

            if(nZeileQuelle > 0)
            {
               nPhase = (nZeileZiel-nZeileQuelle)&0x03;
            }
            else
            {
               // Bei den ersten 32 Zeilen Farbe invertieren. Bei PALplus kommt das 
               // erste Halbbild (+1), sonst das zweite (+2).
               if(m_bPALplus)
                  nPhase = (nZeileZiel+1-nZeileQuelle)&0x03;
               else
                  nPhase = (nZeileZiel+2-nZeileQuelle)&0x03;
            }

            // Wenn der Abstand zu gro wird die Farbe aus der letzten gltigen Zeile holen
            if(nZeileZiel-nZeileQuelle > 40)
               nPhase = 4;
         }
      }

      // Die Bild auf Bildschirmhhe strecken
      while(nScreenLine < nZeileZiel*m_nHoeheZiel/ZEILENZAHL)
      {
         pZiel = (DWORD*)(pDestination+nScreenLine*nBreiteZiel/4);

         if(bDecode == false)
         {
            nPhase       = 0;
            pQuelle      = (DWORD*)pLinesAddress[nZeileZiel];
            pColorQuelle = (DWORD*)pLinesAddress[nLastColorLine];
         }
         else
         {
            pQuelle      = (DWORD*)pLinesAddress[pSortLines[nZeileZiel]];
            pColorQuelle = (DWORD*)pLinesAddress[pSortLines[nLastColorLine]];
         }

         // Farbe korrigieren
         switch(nPhase)
         {
            case 0:
               nLastPhase     = nPhase;
               nLastColorLine = nZeileZiel;

               // v = v, u = u
               for(nPixel=0; nPixel<m_nBreiteQuelle; nPixel+=2)
               {
                  nUV = *pQuelle>>4 & 0xfc0 | *pQuelle>>14 & 0x3f000;

                  nRGB1 = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | nUV];
                  nRGB2 = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | nUV];

                  *pZiel++ = nRGB1 | nRGB2<<16;
                  pQuelle++;
               }
            break;

            case 2:
               nLastPhase     = nPhase;
               nLastColorLine = nZeileZiel;

               // v = -v, u = -u
               for(nPixel=0; nPixel<m_nBreiteQuelle; nPixel+=2)
               {
                  nUV = *pQuelle^0xff00ff00;
                  nUV = nUV>>4 & 0xfc0 | nUV>>14 & 0x3f000;

                  nRGB1 = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | nUV];
                  nRGB2 = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | nUV];

                  *pZiel++ = nRGB1 | nRGB2<<16;
                  pQuelle++;
               }
            break;

            default:
               switch(nLastPhase)
               {
                  case 0:
                     // v = v, u = u
                     for(nPixel=0; nPixel<m_nBreiteQuelle; nPixel+=2)
                     {
                        nUV = *pColorQuelle>>4 & 0xfc0 | *pColorQuelle>>14 & 0x3f000;

                        nRGB1 = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | nUV];
                        nRGB2 = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | nUV];

                        *pZiel++ = nRGB1 | nRGB2<<16;
                        pQuelle++;
                        pColorQuelle++;
                     }
                  break;

                  case 2:
                     // v = -v, u = -u
                     for(nPixel=0; nPixel<m_nBreiteQuelle; nPixel+=2)
                     {
                        nUV = *pColorQuelle^0xff00ff00;
                        nUV = nUV>>4 & 0xfc0 | nUV>>14 & 0x3f000;

                        nRGB1 = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | nUV];
                        nRGB2 = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | nUV];

                        *pZiel++ = nRGB1 | nRGB2<<16;
                        pQuelle++;
                        pColorQuelle++;
                     }
                  break;

                  default:
                     // Alle andern Zeilen ohne Farbe anzeigen
                     for(nPixel=0; nPixel<m_nBreiteQuelle; nPixel+=2)
                     {
                        nRGB1 = m_pYUV2RGBTabelle[*pQuelle>>2 & 0x3f | 0x20800];
                        nRGB2 = m_pYUV2RGBTabelle[*pQuelle>>18 & 0x3f | 0x20800];
                        *pZiel++ = nRGB1 | nRGB2<<16;
                        pQuelle++;
                     }
               }
            break;
         }

         nScreenLine++;
      }
   }
}