#pragma warning(disable: 4056 4201 4514)

#include <windows.h>
#include <stdio.h>

#include "Syster.h"
#include "Debug.h"
#include "resource.h"


/*
 * Ein wenig Theorie
 *
 * Bei Syster/Nagravision werden die Positionen der Zeilen permutiert. Der Decoder hat RAM fr 32
 * Zeilen und da schreibt er auch erstmal die letzten 32 Zeilen eines verschlsselten Halbbildes
 * rein bevor irgendwas anderes passiert. Das sind nmlich die ersten 32 Zeilen des decodierten
 * Halbbildes, was man auch sehr schn erkennt, wenn man sich das verschlsselte Bild mal anschaut.
 * Danach schiebt er Zeile fr Zeile an den Fernseher und holt sich fr jede Zeile die er
 * rausschiebt sofort eine neue. Die Reihenfolge bestimmt, wie bei videocrypt, ein pseudo random
 * number generator, der durch einen verschlsselten seed gestartet wird. Dieser prng ist ein sehr
 * einfaches Modell. Er besteht aus einer Liste von 256 geschickt angeordneten Zahlen zwischen 0
 * und 31. Der seed besteht aus zwei Byte, das erste gibt an, an welcher Stelle in dieser Liste
 * begonnen wird, das zweite, mit welcher ungeraden Schrittweite round robin durch die Liste
 * gegangen wird (deshalb auch 256 x 128 = 32768 mgliche Permutationen). Heraus kommt jeweils
 * eine Zahl zwischen 0 und 31. Dies ist der Index der Zeile im RAM des Decoders, die als nchstes
 * an den Ferseher geschoben wird.
 * 
 * Der Prozessor und die gesammte restliche Hardware wird duch den Zeilentakt des Videosignals
 * getaktet, damit eine hundertprozentige Synkronitt gewhrleistet werden kann. 
 * 
 * Die Decodierungssoftware fr den PC funktioniert nun so:
 * Die Software braucht als Grundlage die oben beschriebene Liste. Es handelt sich einfach nur um
 * eine Datei, die 256 Zeilen enthlt, jede Zeile mit einer Zahl zwischen 0 und 31. Daraus werden
 * nun alle! 32768 Mglichen Permutationen errechnet.
 * 
 * Die Arbeit des Algorithmus besteht nun darin, so schnell wie mglich die richtigen zu finden.
 * Der Algorithmus beruht auf folgendem Prinzip: "Wenn zwei im unverschlsselten Bild
 * aufeinanderfolgende Bildzeilen eines Fields (Halbbildes), die im verschlsselten Bild von
 * einander entfernt liegen, nach einer Transformation mit einer bestimmten Permutation wieder
 * korrekterweise untereinander liegen, dann ist dieser Schlssel mit groer Wahrscheinlichkeit
 * der richtige." Selbstverstndlich sind zwei Zeilen nicht ausreichend um sicher zu sein, das
 * die angenommene Permutation auch die richtige ist. Im Folgenden werde ich diese Zeilenpaare
 * "Kandidaten" nennen. 
 * 
 * Aber wie kann man herausfinden, ob zwei Zeilen zusammen gehren ? Indem man beispielsweise
 * ihre Korrelationsrate berechnet: Wenn die Summe der Betrge der Luminanzdifferenz jedes Pixels
 * der beiden Zeilen "klein" ist, dann liegen sie wahrscheinlich nahe bei einander. Der Erfolg
 * dieser Methode hngt davon ab, da weit voneinander entfernte Zeilen des Bildes relativ
 * unterschiedlich, nahe beieinanderligende aber sehr hnlich sind. Das ist bei Video im 
 * allgemeinen der Fall, auer wenn das Bild sehr einfrmig ist (z.B. ein einfarbiger Hintergrund)
 * oder wenn vertikale Symmetrien vorhanden sind (ein einfarbiges, geometrisches Bild mit
 * senkrechtem Rand). Kurz, je inhaltsreicher das Bild ist, um so besser geht es. Um die
 * Gltigkeit einer Permutation festzustellen, mu man also
 * 
 * 1.) eine grere Anzahl Kandidaten whlen. Je kleiner die Anzahl, um so schneller aber
 *     auch um so unzuverlssiger ist der Agorithmus.
 * 
 * 2.) denjenigen variablen Schlssel bestimmen, der das beste Korrelationsergebnis aller
 *     seiner Kandidaten erhlt (es gengt, die individuellen Korrelationsraten zu summieren).
 * 
 * Die Anzahl der Pixel, die zur Berechnung der Korrelationsrate verwendet werden, kann geringer
 * sein als die Anzahl der Pixel einer Zeile. Dabei ist es sinnvoll, aus verschiedenen Bereichen
 * der Zeile einige nebeneinanderliegende Pixel zu nehmen, um Kontraste in der Senkrechten
 * gnstig auszuwerten. Je mehr Pixel man nimmt, um so zuverlssiger, aber auch um so langsamer
 * arbeitet der Algorithmus. Alle diese Parameter haben eine groe Wirkung auf die Geschwindigkeit
 * des Algorithmus. Um eine Vorstellung davon zu geben: Wir schauen uns 12 Zeilen an. Daraus
 * ergeben sich 800 mglichen Kandidaten (das sind alle mglichen Reihenfolgen dieser 12 Zeilen).
 * In jeder Zeile schauen wir uns jeweils 4 von 16 Pixeln an. D.h. es mssen fr jedes Bild 800
 * Korrelationsraten berechnet werden (800*2*(768*4/16))=307200 Pixel bearbeiten  und dann
 * 32768*12=393216 Mglichkeiten bewertet werden. Das mit 25 multipliziert ergibt den Aufwand
 * fr 1 Sekunde real-time Video. Hinzu kommt (fr 16 bit Farbe): 440k*25 = 11M Transferts vom
 * PCI-Bus zum Speicher durch den BT848 und 11 M Transferts vom Speicher zur Grafikkarte. Das
 * macht 22M/s. Ein AGP-Bus drfte die Last auf dem PCI-Bus halbieren. Die Last auf dem PCI-Bus
 * hat eine nicht vernachlssigbare Wirkung auf die Gte des Algorithmus, was teilweise den
 * Abfall der Bildrate zwischen den 8, 16 und 32-bit Versionen erklrt. 
 */

// Anzahl der Pixel die getestet werden
const int TESTPIXEL        = 24;

// Anzahl der Mglichkeiten
const int KOMBINATIONEN    = 32768;

// Kleinste Testzeile
const int KLEINSTE_ZEILE   = 92;

// Grte Testzeile
const int GROESSTE_ZEILE   = ZEILENZAHL-KLEINSTE_ZEILE-1;

// Name der Keydatei
const char KEY_DATEI[]     = "Key.txt";

// Name der Indexdatei
const char INDEX_DATEI[]   = "MoreTV%02d.idx";

// Relative Position der Bits in der PalPlus Zeile fr eine horizontale Auflsung von 768
const short pPALplusCode[] = {  18,  27,  36,  45,  54,  63,  72,  81,  90, 100, 110, 120, 134, 149};
const short pPALplusData[] = { 160, 178, 196, 214, 232, 250, 268, 286, 304, 322, 340, 358, 376, 394};

// Pegel fr Low und High des PALplus Signals
const short nLevelLow      =  60;
const short nLevelHigh     = 140;


CSyster::CSyster
(
   int   nTestzeilen,
   bool  bSchwarzeZeile,
   bool  bInterpolation
) : m_nTestzeilen(nTestzeilen),
    m_bSchwarzeZeile(bSchwarzeZeile),
   	m_bInterpolation(bInterpolation)
{
   m_bError           = false;
   m_bPALplus         = 0;
   m_nBreiteQuelle    = 0;
   m_nPALplusZeile    = -1;
   m_nPALplusData     = 0;
   m_nSchwarzeZeilen  = 0;
   m_nPaarIndexAnzahl = 0;
   m_pKombinationen   = NULL;
   m_nAnzahlPaare     = 0;
   m_nPaarNummer      = NULL;
   m_nPaarDanach      = NULL;
   m_nPaarDavor       = NULL;
   m_nPaarTreffer     = NULL;

	ZeroMemory(m_pKey, sizeof(m_pKey));
	ZeroMemory(m_nPaarIndex, sizeof(m_nPaarIndex));
	ZeroMemory(m_pSchwarzeZeile, sizeof(m_pSchwarzeZeile));
	ZeroMemory(m_nZeilenPaarKorrelation, sizeof(m_nZeilenPaarKorrelation));

   m_pKombinationen = new short[KOMBINATIONEN*m_nTestzeilen];
   if(m_pKombinationen == NULL)
   {
      MessageBox(IDS_ERR_MEMORY);
      m_bError = true;
      return;
   }

   // Permutationstabelle aus Datei einlesen
   if(LeseKeyDatei() == false)
   {
      MessageBox(IDS_ERR_READ_KEY);
      m_bError = true;
      return;
   }

   // Wenn bereits ein Indexfile existiert, brauchen die Daten nicht mehr 
   // berechnet zu werden
   if(LeseIndexDatei() == true)
      return;

   // IndexTabellen aufbauen
   if(BaueIndex() == false)
   {
      m_bError = true;
      return;
   }

   // Das ganze frs nchste mal abspeichern
   if(SchreibeIndexDatei() == false)
   {
      MessageBox(IDS_ERR_SCHREIBEINDEX);
      m_bError = true;
      return;
   }
}


CSyster::~CSyster()
{
   if(m_pKombinationen != NULL)
      delete[] m_pKombinationen;
}


bool CSyster::FehlerStatus()
{
   return m_bError;
}


bool CSyster::BaueIndex()
{
   // Anzahl der Paare im Unverschlsseltem Zustand 
   // Das heist, die Zeilen sind direkt untereinander
   m_nAnzahlPaare = GROESSTE_ZEILE - KLEINSTE_ZEILE;

   m_nPaarNummer = new int[KOMBINATIONEN];
   if(m_nPaarNummer == NULL)
   {
      MessageBox(IDS_ERR_MEMORY);
      return false;
   }

   // In dieser Tabelle stehen alle Testzeilen einer Kombination hintereinander
   m_nPaarDanach = new int[KOMBINATIONEN*m_nAnzahlPaare];
   if(m_nPaarDanach == NULL)
   {
      MessageBox(IDS_ERR_MEMORY);
      return false;
   }

   m_nPaarDavor = new int[ZEILENZAHL*ZEILENZAHL];
   if(m_nPaarDavor == NULL)
   {
      MessageBox(IDS_ERR_MEMORY);
      return false;
   }

   m_nPaarTreffer = new int[ZEILENZAHL*ZEILENZAHL];
   if(m_nPaarTreffer == NULL)
   {
      MessageBox(IDS_ERR_MEMORY);
      return false;
   }

   int nZeilenIndex;

   // Kennzeichen fr nicht belegt
   for(nZeilenIndex=0; nZeilenIndex<ZEILENZAHL*ZEILENZAHL; nZeilenIndex++)
      m_nPaarDavor[nZeilenIndex] = -1;

   // Anzahl Treffer auf Null setzen
   for(nZeilenIndex=0; nZeilenIndex<ZEILENZAHL*ZEILENZAHL; nZeilenIndex++)
      m_nPaarTreffer[nZeilenIndex] = 0;

   // Eine doppelt verzeigerte Liste aller Zeilenpaare aufbauen
   int nPaarDavorIndex;
   int nPaarDanachIndex;
   int nKey;
   int pZeilen[ZEILENZAHL];
   int nZeileA;
   int nZeileB;

   // Schleife ber alle Kombination
   for(nKey=0; nKey<KOMBINATIONEN; nKey++)
   {
      // Zeilenfolge fr diese Kombination finden
      BerechnePermutation(nKey, pZeilen);

      // Erste Zeile eines Paares
      nZeileA = pZeilen[KLEINSTE_ZEILE];
      for(nZeilenIndex=KLEINSTE_ZEILE+1; nZeilenIndex<=GROESSTE_ZEILE; nZeilenIndex++)
      {
         // Zweite Zeile eines Paars
         nZeileB = pZeilen[nZeilenIndex];

         // Index immer aus "kleine Zeile * Zeilenzahl + groe Zeile" berechnen
         if(nZeileB < nZeileA)
            nPaarDavorIndex = nZeileB*ZEILENZAHL + nZeileA;
         else
            nPaarDavorIndex = nZeileA*ZEILENZAHL + nZeileB;

         // Index auf die Paar danach Tabelle
         nPaarDanachIndex = nKey*m_nAnzahlPaare + nZeilenIndex - (KLEINSTE_ZEILE+1);

         // Zeiger auf das Paar danach
         m_nPaarDanach[nPaarDanachIndex] = m_nPaarDavor[nPaarDavorIndex];
         
         // Zeiger auf das Paar davor
         m_nPaarDavor[nPaarDavorIndex] = nPaarDanachIndex;

         // Anzahl der Treffer von diesem Paars erhhen
         m_nPaarTreffer[nPaarDavorIndex]++;

         // Erste Zeile ist gleich zweiter Zeile
         nZeileA = nZeileB;
      }
   }

   for(nZeilenIndex=0; nZeilenIndex<KOMBINATIONEN; nZeilenIndex++)
      m_nPaarNummer[nZeilenIndex] = 0;

   int nCount = 0;
   int nTreffer;
   int nMaxTreffer;

   m_nPaarIndexAnzahl = 0;
   do
   {
      // Das Paar mit den meisten Treffern suchen
      nMaxTreffer = 0;
      nPaarDavorIndex = 0;
      for(nZeilenIndex=0; nZeilenIndex<ZEILENZAHL*ZEILENZAHL; nZeilenIndex++)
      {
         if(m_nPaarTreffer[nZeilenIndex] > nMaxTreffer)
         {
            nMaxTreffer = m_nPaarTreffer[nZeilenIndex];
            nPaarDavorIndex = nZeilenIndex;
         }
      }

      // Wenn das Paar benutzt wird dann merken
      nTreffer = PaarBewerten(nPaarDavorIndex);
      if(nTreffer > 0)
      {
         nCount += nTreffer;
         PaarMerken(nPaarDavorIndex);
      }

      m_nPaarTreffer[nPaarDavorIndex] = 0;
   }
   while(nCount < KOMBINATIONEN*m_nTestzeilen);

   delete[] m_nPaarNummer;
   delete[] m_nPaarTreffer;
   delete[] m_nPaarDavor;
   delete[] m_nPaarDanach;

   return true;
}


void CSyster::BerechnePermutation
(
   int   nKey,
   int  *pZeilen
)
{
   int   nStartwert = nKey>>7;
   int   nSchrittweite = (nKey&0x7f)*2+1;
   int   nZeile;
   int   nBuffer[32];

   //
   // Diese Funktion berechnet die Zeilenpermutation fr eine Kombination 
   // aus nStartwert und nSchrittweite.
   //
   for(nZeile=0; nZeile<32; nZeile++)
      nBuffer[nZeile] = nZeile;

   for(nZeile=0; nZeile<ZEILENZAHL-32; nZeile++)
   {
      pZeilen[nZeile] = nBuffer[m_pKey[nStartwert]];
      nBuffer[m_pKey[nStartwert]] = nZeile+32;
      nStartwert += nSchrittweite;
      nStartwert &= 0xff;
   }

   for(nZeile=0;nZeile<32;nZeile++)
      pZeilen[nZeile+ZEILENZAHL-32] = nBuffer[nZeile];

   //
   // Interpolieren der ersten 32 Eingangszeilen
   //
   if (m_bInterpolation==true) {
      int i;
      int j;
      int k=ZEILENZAHL>>1;
      for(nZeile=1; nZeile<k; nZeile++)
      {
         i = j = nZeile;
  	     while (pZeilen[nZeile] < 33)
         {
		        if (i>0) i--;
		        if (j<(ZEILENZAHL-1)) j++;
		        if (pZeilen[i]>32)
     		       pZeilen[nZeile]=pZeilen[i];
  		      else if (pZeilen[j]>32)
    		  	   pZeilen[nZeile]=pZeilen[j];
		     }
	    }
   }

}



bool CSyster::FindeZeilenFolge
(
   int **pLinesAddress,
   int  *pSortLines,
   int  *nPALplusData
)
{
   if(m_bPALplus)
   {
      // PalPlus Zeile suchen. Wenn keine Zeile gefunden wurde, abbrechen
      if(FindePALplus(pLinesAddress) == false)
         return false;

      *nPALplusData = m_nPALplusData;
   }

   // Anzahl und Position der schwarzen Zeilen suchen
   if(m_bSchwarzeZeile)
      FindeSchwarzeZeilen(pLinesAddress);

   // Testpixel in einen Zwischenspeicher schreiben. Das erhht die Geschwindigkeit.
   int nPixelSchritt = m_nBreiteQuelle/TESTPIXEL/2;
   int nStartPixel   = nPixelSchritt/2;
   int nTestPixel[ZEILENZAHL][TESTPIXEL];

   for(int nZeile=0; nZeile<ZEILENZAHL; nZeile++)
      for(int nPixel=0; nPixel<TESTPIXEL; nPixel++)
         nTestPixel[nZeile][nPixel] = pLinesAddress[nZeile][nStartPixel+nPixel*nPixelSchritt] & 0xff;

   // Die Korrelation zwischen den Zeilen berrechnen
   int nZeileA;
   int nZeileB;
   int nKorrelationZeile;
   int nKorrelationPixel;

   for(int nIndex=0; nIndex<m_nPaarIndexAnzahl; nIndex++)
   {
      nZeileA = m_nPaarIndex[nIndex] >> 9;
      nZeileB = m_nPaarIndex[nIndex] & 0x1ff;

      nKorrelationZeile = 0;
      for(int nPixel=0; nPixel<TESTPIXEL; nPixel++)
      {
         nKorrelationPixel = nTestPixel[nZeileA][nPixel] - nTestPixel[nZeileB][nPixel];

         // Ein kleiner Ersatz fr die Funktion abs()
         if(nKorrelationPixel < 0)
            nKorrelationZeile -= nKorrelationPixel;
         else
            nKorrelationZeile += nKorrelationPixel;
      }

      m_nZeilenPaarKorrelation[nIndex] = nKorrelationZeile;
   }

   // Key bestimmen und Permutation berechnen
   BerechnePermutation(SucheKey(), pSortLines);

   return true;
}



bool CSyster::FindePALplus
(
   int  **pLinesAddress
)
{
   int    nBit;
   int    nPixel;
   int    nZeile;
   int    nOffset;
   short *pLine;

   // Daten initialisieren
   m_nPALplusZeile = -1;
   m_nPALplusData  =  0;

   // Die 32 mglichen Zeilen durchsuchen
   for(nZeile=0; nZeile<32; nZeile++)
   {  
      pLine = (short*)&pLinesAddress[nZeile][0];

      // Auf das Startbit warten
      nOffset = 0;
      for(nPixel=0; nPixel<20*m_nBreiteQuelle/768; nPixel++)
      {
         // Startbit gefunden
         if((*(pLine+nOffset)&0xff) > nLevelHigh)
            break;
         
         // Abstand Zeilenanfang bis Startbit
         nOffset++;
      }

      // Bitcode berprfen
      for(nPixel=0; nPixel<14; nPixel+=2)
      {
         // Fr andere horizontale Auflsungen als 768, Pixelabstand angleichen
         if((*(pLine+nOffset+pPALplusCode[nPixel]*m_nBreiteQuelle/768)&0xff) > nLevelLow)
            break;

         if((*(pLine+nOffset+pPALplusCode[nPixel+1]*m_nBreiteQuelle/768)&0xff) < nLevelHigh)
            break;
      }

      // Wenn alle 7 Bit richtig sind dann wurde die Zeile gefunden
      if(nPixel == 14)
      {  
         m_nPALplusZeile = nZeile;

         // Daten auslesen
         nBit = 0x0001;

         for(nPixel=0; nPixel<14; nPixel++)
         {
            if((*(pLine+nOffset+pPALplusData[nPixel]*m_nBreiteQuelle/768)&0xff) > nLevelHigh)
               m_nPALplusData |= nBit;

            nBit <<= 1;
         }
         
         return true;
      }
   }

   // Zeile nicht gefunden
   return false;
}

  
void CSyster::FindeSchwarzeZeilen
(
   int    **pLinesAddress
)
{
   int      nZeile;
   int      nPixel;
   int      nPixelSchritt = m_nBreiteQuelle/TESTPIXEL/2;
   int      nStartPixel   = nPixelSchritt/2;
   short   *pLine;

   m_nSchwarzeZeilen = 0;

   // Die 32 mglichen Zeilen durchsuchen
   for(nZeile=0; nZeile<32; nZeile++)
   {
      // Adresse des ersten Pixel
      pLine = (short*)&pLinesAddress[nZeile][nStartPixel];

      // Pixel testen
      for(nPixel=0; nPixel<TESTPIXEL; nPixel++)
      {
         // Wenn ein Pixel das Limit berschreitet, ist die Zeile nicht schwarz
         if((*(pLine+nPixelSchritt*nPixel)&0xff) > nLevelLow)
            break;
      }

      // Merken ob die Zeile schwarz ist
      if(nPixel == TESTPIXEL)
      {
         m_nSchwarzeZeilen++;
         m_pSchwarzeZeile[nZeile] = true;
      }
      else
      {
         m_pSchwarzeZeile[nZeile] = false;
      }
   }
}


bool CSyster::LeseIndexDatei()
{
   FILE *sFile;

   // Datei mit vorberechneten Tabellen einlesen
   char pDateiName[20];
   
   // Dateiname erzeugen
   sprintf(pDateiName, INDEX_DATEI, m_nTestzeilen);

   // Datei zum Lesen ffnen
   sFile = fopen(pDateiName, "rb");

   // Wenn die Datei nicht vorhanden ist, Tabellen neu berechnen
   if(sFile == NULL)
      return false;

   // Daten einlesen      
   if(fread(m_pKombinationen, sizeof(short), KOMBINATIONEN*m_nTestzeilen, sFile) != (size_t)(KOMBINATIONEN*m_nTestzeilen))
   {
      fclose(sFile);
      return false;
   }

   if(fread(m_nPaarIndex, sizeof(int), INDEXSIZE, sFile) != INDEXSIZE)
   {
      fclose(sFile);
      return false;
   }

   if(fread(&m_nPaarIndexAnzahl, sizeof(int), 1, sFile) != 1)
   {
      fclose(sFile);
      return false;
   }

   // Daten wurden erfolgreich eingelesen
   fclose(sFile);
   return true;
}


bool CSyster::LeseKeyDatei()
{
   // Permutationstabelle aus Datei einlesen
   int   nZlr;
   int   nInputValue;
   FILE *sFile = fopen(KEY_DATEI, "r");

   if(sFile == NULL)
      return false;

   for(nZlr=0; nZlr<256; nZlr++)
   {
      if(fscanf(sFile, "%d\n", &nInputValue) == EOF)
         return false;

      m_pKey[nZlr] = (char)(nInputValue&0x1f);
   }

   fclose(sFile);
   return true;
}


int CSyster::PaarBewerten
(
   int nPaarDavorIndex
)
{
   int nKey;
   int nTreffer = 0;
   int nNaechstesPaarIndex = m_nPaarDavor[nPaarDavorIndex];

   while(nNaechstesPaarIndex != -1)
   {
      nKey = nNaechstesPaarIndex/m_nAnzahlPaare;

      if(m_nPaarNummer[nKey] < m_nTestzeilen)
         nTreffer++;

      nNaechstesPaarIndex = m_nPaarDanach[nNaechstesPaarIndex];
   }

   return nTreffer;
}


void CSyster::PaarMerken
(
   int nPaarDavorIndex
)
{
   int nKey;
   int nNaechstesPaar;
   int nNaechstesPaarIndex;
   int nZeileA;
   int nZeileB;
   int nNummer;

   nZeileA = nPaarDavorIndex / ZEILENZAHL;
   nZeileB = nPaarDavorIndex % ZEILENZAHL;
   m_nPaarIndex[m_nPaarIndexAnzahl] = nZeileA<<9 | nZeileB;

   nNaechstesPaarIndex = m_nPaarDavor[nPaarDavorIndex];
   while(nNaechstesPaarIndex != -1)
   {
      nKey = nNaechstesPaarIndex/m_nAnzahlPaare;
      nNummer = m_nPaarNummer[nKey]++;
      nNaechstesPaar = m_nPaarDanach[nNaechstesPaarIndex];

      if(nNummer < m_nTestzeilen)
         m_pKombinationen[m_nTestzeilen*nKey+nNummer] = (short)m_nPaarIndexAnzahl;

      nNaechstesPaarIndex = nNaechstesPaar;
   }

   m_nPaarIndexAnzahl++;
}


bool CSyster::SchreibeIndexDatei()
{
   FILE *sFile;

   // Vorberechneten Tabellen in Datei schreiben
   char pDateiName[20];
   
   // Dateiname erzeugen
   sprintf(pDateiName, INDEX_DATEI, m_nTestzeilen);

   // Datei zum Schreiben ffnen
   sFile = fopen(pDateiName, "wb");

   if(sFile == NULL)
      return false;

   if(fwrite(m_pKombinationen, sizeof(short), KOMBINATIONEN*m_nTestzeilen, sFile) != (size_t)(KOMBINATIONEN*m_nTestzeilen))
   {
      fclose(sFile);
      return false;
   }

   if(fwrite(m_nPaarIndex, sizeof(int), INDEXSIZE, sFile) != INDEXSIZE)
   {
      fclose(sFile);
      return false;
   }

   if(fwrite(&m_nPaarIndexAnzahl, sizeof(int), 1, sFile) != 1)
   {
      fclose(sFile);
      return false;
   }

   fclose(sFile);
   return true;
}


//--- Espresso's nderungen

int CSyster::SucheKey()
{
   int      nZeilenZlr;
   int      nKorrelation;
   int      nBesterStartwert   = 0;
   int      nBesteSchrittweite = 0;
   int      nBesteKorrelation  = 0x7FFFFFFF;
   short   *pKombinationen     = m_pKombinationen;

   int      nPermutationsZlr;
   int      nBesterKandidat   = 0;

   // Prfen welche Permutation die Beste ist.
   for(nPermutationsZlr=0; nPermutationsZlr<KOMBINATIONEN; nPermutationsZlr++)
   {
	  nKorrelation =  *((unsigned int*)m_nZeilenPaarKorrelation+*pKombinationen);
	  nKorrelation += *((unsigned int*)m_nZeilenPaarKorrelation+*(pKombinationen+1));
	  nKorrelation += *((unsigned int*)m_nZeilenPaarKorrelation+*(pKombinationen+2));
	  nKorrelation += *((unsigned int*)m_nZeilenPaarKorrelation+*(pKombinationen+3));
	  nKorrelation += *((unsigned int*)m_nZeilenPaarKorrelation+*(pKombinationen+4));
	  nKorrelation += *((unsigned int*)m_nZeilenPaarKorrelation+*(pKombinationen+5));
	  nKorrelation += *((unsigned int*)m_nZeilenPaarKorrelation+*(pKombinationen+6));
      nZeilenZlr=(int)7;
Loop2: 
	  nKorrelation += *((unsigned int*)m_nZeilenPaarKorrelation+*(pKombinationen+nZeilenZlr++));
	  if(nKorrelation >= nBesteKorrelation)
         {
			goto InnerLoopOut ;
         }
	  if ( nZeilenZlr >= m_nTestzeilen ) goto Out2;
	  goto Loop2  ;
Out2:

      if(nKorrelation < nBesteKorrelation)
      {
         nBesteKorrelation = nKorrelation;
         nBesterKandidat = nPermutationsZlr;
      }
InnerLoopOut: ;
      pKombinationen+=m_nTestzeilen;
   }
   return nBesterKandidat;
}
