//----------------------------------//
// VideoCapture.c   by:   KennyWood //
//----------------------------------//

#include "Capture.h"
#include "Decoder.h"
#include "BTChip.h"

//-----------------------------------------------------------------------------

#define MAXDRIVERS       5
#define MAXDNAMELENGTH 128
#define MAXDVERSLENGTH  32

//-----------------------------------------------------------------------------

HGLOBAL hDIBmem ;

//-----------------------------------------------------------------------------

struct
{
	char name    [ MAXDNAMELENGTH ] ;
	char version [ MAXDVERSLENGTH ] ;
} driver [ MAXDRIVERS ] ;

int numDrivers ;

//-----------------------------------------------------------------------------
//
// VideoCapture_FindDrivers:
//
//  Let's check which drivers have we got.
//
void VideoCapture_FindDrivers ()
{
	numDrivers = 0 ;

	while ( numDrivers < MAXDRIVERS )
	{
		if ( !capGetDriverDescription( numDrivers , 
									   driver[numDrivers].name    , MAXDNAMELENGTH , 
									   driver[numDrivers].version , MAXDVERSLENGTH ) )
			break ;
		numDrivers ++ ;
	}

	haveBT848 = BTChip_Init() ;
}

//-----------------------------------------------------------------------------
//
// VideoCapture_GetDriverName:
//
//  Let's check which drivers have we got.
//
void VideoCapture_GetDriverName ( int i , char* str )
{
	sprintf( str , "[%d] %s %s" , i , driver[i].name , driver[i].version ) ;
}

//-----------------------------------------------------------------------------
//
// VideoCapture_Init:
//
//  Here we make some setup for our video capturer.
//
BOOL VideoCapture_Init ( int driver )
{
	if( !capDriverConnect( hCapWnd , driver ) )
		return FALSE ;

	capOverlay( hCapWnd , FALSE ) ;
	capPreview( hCapWnd , FALSE ) ;

	return TRUE ;
}

//-----------------------------------------------------------------------------
//
// VideoCapture_SetVideoFormat:
//
//  Here we set up the dimensions for the acquired data.
//
BOOL VideoCapture_SetVideoFormat ( int width , int height , int bpp )
{
	const DWORD VIDEO_FORMAT = MAKEFOURCC('Y','U','Y','2') ;

	BITMAPINFO bi ;
	bi.bmiHeader.biSize          = sizeof(BITMAPINFOHEADER) ;
	bi.bmiHeader.biWidth         = width ;
	bi.bmiHeader.biHeight        = height ;
	bi.bmiHeader.biBitCount      = bpp ;
	bi.bmiHeader.biCompression   = VIDEO_FORMAT ;
	bi.bmiHeader.biPlanes        = 1 ;
	bi.bmiHeader.biSizeImage     = (width * height * bpp) / 8 ;
	bi.bmiHeader.biClrUsed       = 0 ;
	bi.bmiHeader.biXPelsPerMeter = 0 ;
	bi.bmiHeader.biYPelsPerMeter = 0 ;
	bi.bmiHeader.biClrImportant  = 0 ;
	return capSetVideoFormat( hCapWnd , &bi , sizeof(bi) ) ;
}

//-----------------------------------------------------------------------------
//
// VideoCapture_Open:
//
//  Here we make some setup for the characteristics of the captured frames.
//
BOOL VideoCapture_Open ( int width , int height , int depth )
{
	DWORD        rgb16masks [ 4 ] = { 0xF800 , 0x07E0 , 0x001F , 0x0000 } ;
	CAPTUREPARMS capParms ;
	char		 text [ 256 ] ;
	
	// Set up video format...
	if ( !VideoCapture_SetVideoFormat( width , height , depth ) )
	{
		sprintf( text , "Can't set capture at %dx%dx%d" , width , height , depth ) ;
		MessageBox( hMainWnd , text , appName , MB_ICONEXCLAMATION ) ;
		return FALSE ;
	}

	// Set up capture format...
	capSetCallbackOnVideoStream( hCapWnd , VideoStreamCallback ) ;
	capCaptureGetSetup( hCapWnd, &capParms , sizeof(CAPTUREPARMS) ) ;
	capParms.dwRequestMicroSecPerFrame = 1000000/25 ;
	capParms.fMakeUserHitOKToCapture   = FALSE ;
	capParms.fAbortLeftMouse           = FALSE ;
	capParms.fAbortRightMouse          = FALSE ;
	capParms.fYield                    = TRUE ;
	capParms.fCaptureAudio             = FALSE ;
	capParms.wPercentDropForError      = 10 ;
	capParms.wNumVideoRequested        = 3 ;
	capParms.wNumAudioRequested        = 0 ;
	if ( !capCaptureSetSetup( hCapWnd , &capParms , sizeof(CAPTUREPARMS) ) )
	{
		MessageBox( hMainWnd , "Can't set video setup" , appName , MB_ICONEXCLAMATION ) ;
		return FALSE ;
	}
	
	hDIBmem = GlobalAlloc( GHND , sizeof(BITMAPINFOHEADER) + sizeof(rgb16masks) ) ;
	DIB = GlobalLock( hDIBmem ) ;
	
	// Set up DIB for captured frames...
	DIB -> bmiHeader.biSize          = sizeof(DIB -> bmiHeader) ;
	DIB -> bmiHeader.biWidth         = width ;
	DIB -> bmiHeader.biHeight        = BIFIELDHEIGHT + 1 ;
	DIB -> bmiHeader.biPlanes        = 1 ;
	DIB -> bmiHeader.biBitCount      = depth ;
	DIB -> bmiHeader.biCompression   = BI_BITFIELDS ;
	DIB -> bmiHeader.biSizeImage     = DIB -> bmiHeader.biWidth *
									   DIB -> bmiHeader.biHeight *
									   DIB -> bmiHeader.biBitCount / 8 ;
	DIB -> bmiHeader.biXPelsPerMeter = 0 ;
	DIB -> bmiHeader.biYPelsPerMeter = 0 ;
	DIB -> bmiHeader.biClrUsed       = 0 ;
	DIB -> bmiHeader.biClrImportant  = 0 ;
	memcpy( &(DIB -> bmiColors) , rgb16masks , sizeof(rgb16masks) ) ;

	return TRUE ;
}

//-----------------------------------------------------------------------------
//
// VideoCapture_Close:
//
//  Here we shut down the video capturer.
//
void VideoCapture_Close ()
{
	VideoCapture_Stop() ;
	capSetCallbackOnVideoStream( hCapWnd , NULL ) ;
	capDriverDisconnect( hCapWnd ) ;

	if ( hDIBmem )
	{
		GlobalUnlock( hDIBmem ) ;
		GlobalFree( hDIBmem ) ;
		hDIBmem = NULL ;
	}

	if ( haveBT848 )
		BTChip_Close() ;
}

//-----------------------------------------------------------------------------
//
// VideoCapture_Play/Stop:
//
//  SAFELY starts/stops the video capture.
//
void VideoCapture_Play ()
{
	capCaptureSequenceNoFile( hCapWnd ) ;
}

void VideoCapture_Stop ()
{
	volatile CAPSTATUS status ;

	// We STOP the video capture, so the videostream callback won't be called again
	capCaptureStop( hCapWnd ) ;

	// The callback funcion could be executing at this point, 
	// so we ***HAVE*** to ensure that it has finished to decode
	// the last frame.  That was the cause of the *HUGE* old bug
	// when closing the application in fullScreen mode.
	do
	{		
		capGetStatus( hCapWnd , &status , sizeof(status) ) ;
	}
	while ( status.fCapturingNow ) ;

	Sleep( 500 ) ;   // For Sanity
}
