//---------------------------------------------------------------------------
// FileName     : ioclass.cpp
// Created      : 19/11/98
// Project      : Windows kernel device driver for hardware IO
// Author       : Elli
// Note         : This program is free software; you can redistribute it
//                and/or modify it under the terms of the GNU General Public
//                License as published by the Free Software Foundation;
//                either version 2 of the License, or (at your option) any
//                later version.
//
//                This program is distributed in the hope that it will be useful,
//                but WITHOUT ANY WARRANTY; without even the implied warranty of
//                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//                GNU General Public License for more details.
//
//                You should have received a copy of the GNU General Public
//                License along with this program; if not, write to the Free
//                Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// History      :
//                V 1.00 initial version
//                  1.01 alloc and free memory added
//---------------------------------------------------------------------------

#include "ioclass.h"
#include "debugout.h"


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
CIOAccessDevice::CIOAccessDevice(void)
 {
  ulMemoryBase         = 0;
  ulMappedMemoryLength = 0;    // mapped memory base
  ulBusNumber          = 0;
  ulSlotNumber         = 0;
  virtualAddress       = 0;
  ulMemoryBaseEx       = 0;

  memset( &memoryList, 0, sizeof( memoryList ));

  #if defined (_NTKERNEL_)
   if ( virtualAddress )
    {
     ZwUnmapViewOfSection( (HANDLE) -1, virtualAddress);
    }
  #endif
 }

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
CIOAccessDevice::~CIOAccessDevice()
 {
  if ( ulMemoryBase )
   {
    unmapMemory();
   }

  ULONG       ulIndex;
  PMemoryNode node;

  //
  // Free allocated memory
  //

  for ( ulIndex = 0; ulIndex < MAX_FREE_MEMORY_NODES; ulIndex++)
   {
    node = &memoryList[ ulIndex ];

    if ( node->ulSystemAddress )
     {
      freeMemory ( node );
     }
   }
 }


#if defined (_NTKERNEL_)
//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
NTSTATUS
CIOAccessDevice::deviceIOControl(IN PIRP   irp)
 {
  PIO_STACK_LOCATION      irpStack;
  PVOID                   inputBuffer;
  PVOID                   outputBuffer;
  ULONG                   inputBufferLength;
  ULONG                   outputBufferLength;
  NTSTATUS                ntStatus;
  ULONG                   ioControlCode;

  //
  // Get a pointer to the current location in the Irp. This is where
  //     the function codes and parameters are located.
  //

  irpStack        = IoGetCurrentIrpStackLocation (irp);
  ioControlCode   = irpStack->Parameters.DeviceIoControl.IoControlCode;

  //
  // Get the pointer to the input/output buffer and its length
  //

  switch ( IOCTL_TRANSFER_TYPE(ioControlCode) )
   {
    case METHOD_IN_DIRECT:
      if ( irp->MdlAddress )
       {
        inputBuffer       = MmGetSystemAddressForMdl( irp->MdlAddress );
       }
      else
       {
       inputBuffer        = irp->AssociatedIrp.SystemBuffer;
       }

      inputBufferLength   = irpStack->Parameters.DeviceIoControl.InputBufferLength;
      outputBuffer        = irp->AssociatedIrp.SystemBuffer;
      outputBufferLength  = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
      break;

    case METHOD_OUT_DIRECT:
      if ( irp->MdlAddress )
       {
        outputBuffer      = MmGetSystemAddressForMdl( irp->MdlAddress );
       }
      else
       {
        debugOut(dbError,"mdl for ioControl %X expected",ioControlCode);
       }

      inputBuffer         = irp->AssociatedIrp.SystemBuffer;
      inputBufferLength   = irpStack->Parameters.DeviceIoControl.InputBufferLength;
      outputBufferLength  = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
      break;

    default:
       inputBuffer        = irp->AssociatedIrp.SystemBuffer;
       inputBufferLength  = irpStack->Parameters.DeviceIoControl.InputBufferLength;
       outputBuffer       = irp->AssociatedIrp.SystemBuffer;
       outputBufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
       break;
   }

  ntStatus = deviceControl(
                           ioControlCode,
                           (PHWIOParam) inputBuffer,
                           (PULONG) outputBuffer,
                           &irp->IoStatus.Information
                          );



  return ntStatus;
 }

#elif defined (WIN95)


NTSTATUS
CIOAccessDevice::deviceIOControl(IN PIRP  irp)
 {
  NTSTATUS ntStatus;
  ULONG    ulBytesWritten;

  switch ( irp->dwIoControlCode )
   {
    case DIOC_OPEN:
      debugOut(dbTrace,"DIOC_OPEN");
      ntStatus = STATUS_SUCCESS;
      break;

    case DIOC_CLOSEHANDLE:
      debugOut(dbTrace,"DIOC_CLOSEHANDLE");
      ntStatus = STATUS_SUCCESS;
      break;

    default:
      ntStatus = deviceControl(
                               irp->dwIoControlCode,
                               (PHWIOParam) irp->lpvInBuffer,
                               (PULONG) irp->lpvOutBuffer,
                               &ulBytesWritten
                              );
      break;
   }

#if defined (later)
  // check this later, this crashs the system

  if ( irp->lpcbBytesReturned )
   {
    *((PULONG)irp->lpcbBytesReturned) = ulBytesWritten;
   }

  //
  // The internal member of overlapped structure contains
  // a pointer to the event structure that will be signalled,
  // resuming the execution of the waiting GetOverlappedResult
  // call.
  //

  if ( irp->lpoOverlapped )
   {
    ((OVERLAPPED *)irp->lpoOverlapped)->O_InternalHigh = ulBytesWritten;
   }
#endif

  return ntStatus;
 }

#endif


//---------------------------------------------------------------------------
// The READ/WRITE_PORT_* calls manipulate I/O registers in PORT space.
// (Use x86 in/out instructions.)
//
// The READ/WRITE_REGISTER_* calls manipulate I/O registers in MEMORY space.
// (Use x86 move instructions)
//---------------------------------------------------------------------------
NTSTATUS
CIOAccessDevice::deviceControl(
                               IN  ULONG       ioControlCode,
                               IN  PHWIOParam  ioParam,
                               OUT PULONG      outputBuffer,
                               OUT PULONG      pulBytesWritten
                              )
 {
  NTSTATUS status;

  status           = STATUS_SUCCESS;
  *pulBytesWritten = 0;

  switch ( ioControlCode )
   {
    case ioctlReadBYTE:
      osPortReadByte( ioParam->dwAddress, *outputBuffer);
      break;

    case ioctlReadWORD:
      osPortReadWord( ioParam->dwAddress, *outputBuffer);
      break;

    case ioctlReadDWORD:
      osPortReadLong( ioParam->dwAddress, *outputBuffer);
      break;

    case ioctlWriteBYTE:
      osPortWriteByte( ioParam->dwAddress, ioParam->dwValue);
      *pulBytesWritten = 1;
      break;

    case ioctlWriteWORD:
      osPortWriteWord( ioParam->dwAddress, ioParam->dwValue);
      *pulBytesWritten = 2;
      break;

    case ioctlWriteDWORD:
      osPortWriteLong( ioParam->dwAddress, ioParam->dwValue);
      *pulBytesWritten = 4;
      break;

    case ioctlMapMemory:
      *outputBuffer = mapMemory( ioParam->dwAddress, ioParam->dwValue);
      *pulBytesWritten = 4;
      break;

    case ioctlUnmapMemory:
      unmapMemory();
      break;

    case ioctlReadMappedMemory:
      if ( ulMemoryBase )
       {
        ULONG ulAddress = ioParam->dwAddress + ulMemoryBase;
        osMemoryRead( ulAddress, *outputBuffer);
        *pulBytesWritten = 4;
        debugOut(dbTrace,"memory %X read %X",ioParam->dwAddress, *outputBuffer);
       }
      break;

    case ioctlWriteMappedMemory:
      if ( ulMemoryBase )
       {
        ULONG ulAddress = ioParam->dwAddress + ulMemoryBase;
        osMemoryWrite( ulAddress, ioParam->dwValue );
        debugOut(dbTrace,"memory %X write %X",ioParam->dwAddress, ioParam->dwValue);
       }
      break;

    case ioctlGetPCIInfo:
      status = pciFindDevice(
                             (USHORT) ioParam->dwAddress,
                             (USHORT) ioParam->dwValue,
                             &ulBusNumber,
                             &ulSlotNumber
                            );

      if ( status == STATUS_SUCCESS)
       {
        if ( ! isValidAddress(outputBuffer ) )
         {
          debugOut(dbError,"! invalid system address %X",outputBuffer);
         }
        else
         {
          status = pciGetDeviceConfig(
                                      ulBusNumber,
                                      ulSlotNumber,
                                      (PPCI_COMMON_CONFIG) outputBuffer
                                    );
         }
       }
      else
       {
        debugOut(dbTrace,"pci device for vendor %lX deviceID %lX not found",ioParam->dwAddress,ioParam->dwValue);
       }

      *pulBytesWritten = sizeof(PCI_COMMON_CONFIG);
      break;

    case ioctlMapMemoryEx:
      *outputBuffer    = mapMemoryEx( ioParam->dwAddress, ioParam->dwValue);
      *pulBytesWritten = 4;
      break;

    case ioctlAllocMemory:
      status           = allocMemory(
                                      ioParam->dwAddress,
                                      outputBuffer,
                                      outputBuffer + 1
                                    );

      *pulBytesWritten = sizeof(DWORD) * 2;
      break;

    case ioctlFreeMemory:
      status           = freeMemory(ioParam->dwAddress);
      *pulBytesWritten = 0;
      break;

    default:
      debugOut(dbError,"unknown command %lX",ioControlCode);
      break;
   }

  return status;
 }




//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
int
CIOAccessDevice::isValidAddress(IN void * pvAddress)
 {
  #if defined (_NTKERNEL_)
    return MmIsAddressValid( pvAddress);
  #else
    return 1;
  #endif
 }


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
ULONG
CIOAccessDevice::mapMemory(ULONG ulPhysicalAddress, ULONG ulLength)
 {
  if ( ! ulMemoryBase )
   {
    ulMappedMemoryLength = ulLength;

    debugOut(dbTrace,"mapMemory for %X", ulPhysicalAddress);

    #if defined (_NTKERNEL_)
      PHYSICAL_ADDRESS translatedAddress;
      PHYSICAL_ADDRESS busAddress;
      ULONG            addressSpace;
      BOOLEAN          bTranslate;

      KIRQL irql;

      KeRaiseIrql(PASSIVE_LEVEL, &irql);

      busAddress.LowPart          = ulPhysicalAddress | 8;
      busAddress.HighPart         = 0;
      translatedAddress.LowPart   = 0;
      translatedAddress.HighPart  = 0;
      addressSpace                = 0xFF;

      bTranslate = HalTranslateBusAddress(
                                           PCIBus,
                                           ulBusNumber,
                                           busAddress,
                                           &addressSpace,
                                           &translatedAddress
                                          );

     if ( ! bTranslate )
      {
       debugOut(dbError,"HalTranslateBusAddress() failed, addressSpace %X",addressSpace);
       translatedAddress.LowPart = ulPhysicalAddress;
      }
     //else
      {
       //
       // memory space
       //

       ulMemoryBase = (ULONG) MmMapIoSpace(translatedAddress, ulLength, MmNonCached);

       debugOut(dbTrace,"map pysical address %X to memory base %X,length %d",translatedAddress.LowPart, ulMemoryBase,ulLength);
      }

     KeLowerIrql(irql);

    #elif (WIN95)

     ulMemoryBase = (ULONG) _MapPhysToLinear( (void *)ulPhysicalAddress, ulLength, 0);

    #endif
   }
  else
   {
    debugOut(dbError,"! mapMemory failed, already mapped to %lX",ulMemoryBase);
    return 0;
   }


  return ulMemoryBase;
 }


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
CIOAccessDevice::unmapMemory(void)
 {
  if ( ulMemoryBase )
   {
    #if defined (_NTKERNEL_)

      debugOut(dbTrace,"MMUnmapIoSpace() %X",ulMemoryBase);

      MmUnmapIoSpace( (PVOID) ulMemoryBase, ulMappedMemoryLength);

    #elif (WIN95)


    #endif

    ulMemoryBase = 0;
  }
 }

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
ULONG
CIOAccessDevice::mapMemoryEx(ULONG ulPhysicalAddress, ULONG ulLength)
 {
  debugOut(dbTrace,"map physical memory %X to user mode",ulPhysicalAddress);

  if ( ulMemoryBaseEx )
   {
    debugOut(dbError,"! mapMemory failed, already mapped to %lX",ulMemoryBaseEx);
    return 0;
   }

 #if defined (WIN95)
 #else

  PHYSICAL_ADDRESS   physicalAddress;
  UNICODE_STRING     physicalMemoryUnicodeString;
  OBJECT_ATTRIBUTES  objectAttributes;
  HANDLE             physicalMemoryHandle  = NULL;
  PVOID              physicalMemorySection = NULL;
  ULONG              inIoSpace, inIoSpace2;
  NTSTATUS           ntStatus;
  PHYSICAL_ADDRESS   tmpPhysicalAddressBase;
  PHYSICAL_ADDRESS   physicalAddressEnd;
  PHYSICAL_ADDRESS   tmpPhysicalAddressEnd;
  PHYSICAL_ADDRESS   viewBase;
  PHYSICAL_ADDRESS   mappedLength;
  BOOLEAN            bTranslateBaseAddress;
  BOOLEAN            bTranslateEndAddress;



  ulMappedMemoryLengthEx   = ulLength;
  physicalAddress.LowPart  = ulPhysicalAddress;
  physicalAddress.HighPart = 0;


  //
  // Get a pointer to physical memory...
  //
  // - Create the name
  // - Initialize the data to find the object
  // - Open a handle to the oject and check the status
  // - Get a pointer to the object
  // - Free the handle
  //

  RtlInitUnicodeString (&physicalMemoryUnicodeString,L"\\Device\\PhysicalMemory");

  InitializeObjectAttributes (
                               &objectAttributes,
                               &physicalMemoryUnicodeString,
                               OBJ_CASE_INSENSITIVE,
                               (HANDLE) NULL,
                               (PSECURITY_DESCRIPTOR) NULL
                             );

  ntStatus = ZwOpenSection (
                            &physicalMemoryHandle,
                            SECTION_ALL_ACCESS,
                            &objectAttributes
                           );


  if ( ! NT_SUCCESS(ntStatus))
   {
    debugOut(dbError,"! ZwOpenSection failed %X",ntStatus);
   }
  else
   {
    ntStatus = ObReferenceObjectByHandle(
                                         physicalMemoryHandle,
                                         SECTION_ALL_ACCESS,
                                         (POBJECT_TYPE) NULL,
                                         KernelMode,
                                         &physicalMemorySection,
                                         (POBJECT_HANDLE_INFORMATION) NULL
                                        );

    if ( ! NT_SUCCESS(ntStatus))
     {
      debugOut(dbError,"! ObReferenceObjectByHandle failed %X",ntStatus);
     }
    else
     {
      //
      // Initialize the physical addresses that will be translated
      //

      tmpPhysicalAddressBase.LowPart = 0;
      tmpPhysicalAddressEnd.LowPart  = 0;
      physicalAddressEnd             = RtlLargeIntegerAdd(
                                                           physicalAddress,
                                                           RtlConvertUlongToLargeInteger( ulLength)
                                                          );

      //
      // Translate the physical addresses.
      //

      debugOut(dbTrace,"translate bus addresses (%X : %X), bus no. %u",physicalAddress.LowPart,physicalAddressEnd.LowPart,ulBusNumber);

      bTranslateBaseAddress = HalTranslateBusAddress (
                                                      PCIBus,
                                                      ulBusNumber,
                                                      physicalAddress,
                                                      &inIoSpace,
                                                      &tmpPhysicalAddressBase
                                                     );

      bTranslateEndAddress =  HalTranslateBusAddress (
                                                      PCIBus,
                                                      ulBusNumber,
                                                      physicalAddressEnd,
                                                      &inIoSpace2,
                                                      &tmpPhysicalAddressEnd
                                                    );


      if ( ! bTranslateBaseAddress ) debugOut(dbError,"cannot translate bus address");
      if ( ! bTranslateEndAddress )  debugOut(dbError,"cannot translate bus end address");

      debugOut(dbTrace,"physical base %X  end %X",tmpPhysicalAddressBase.LowPart,tmpPhysicalAddressEnd.LowPart);

      // if ( bTranslateBaseAddress && bTranslateEndAddress )
       {
        //
        // Calculate the length of the memory to be mapped
        //

        mappedLength = RtlLargeIntegerSubtract (tmpPhysicalAddressEnd,tmpPhysicalAddressBase);

        //
        // If the mappedlength is zero, somthing very weird happened in the HAL
        // since the Length was checked against zero.
        //

        if (mappedLength.LowPart == 0)
         {
          debugOut(dbError,"mappedLength.LowPart == 0");
          ntStatus = STATUS_UNSUCCESSFUL;
         }
        else
         {
          ulLength = mappedLength.LowPart;

          //
          // If the address is in io space, just return the address, otherwise
          // go through the mapping mechanism
          //

          if (inIoSpace)
           {
            ulMemoryBaseEx = tmpPhysicalAddressBase.LowPart;
            debugOut(dbTrace,"mapped to %X, in IO space",ulMemoryBaseEx);
           }
          else
           {

            //
            // initialize view base that will receive the physical mapped
            // address after the MapViewOfSection call.
            //

            viewBase = tmpPhysicalAddressBase;


            //
            // Let ZwMapViewOfSection pick an address
            //

            virtualAddress = NULL;

            //
            // Map the section
            //

            ntStatus = ZwMapViewOfSection (
                                           physicalMemoryHandle,
                                          (HANDLE) -1,
                                          &virtualAddress,
                                          0L,
                                          ulLength,
                                          &viewBase,
                                          &ulLength,
                                          ViewShare,
                                          0,
                                          PAGE_READWRITE | PAGE_NOCACHE
                                         );

           if ( NT_SUCCESS(ntStatus))
            {
             debugOut(dbError,"! ZwMapViewOfSection failed %X",ntStatus);
            }
           else
            {
             //
             // Mapping the section above rounded the physical address
             // down to the nearest 64 K boundary. Now return a virtual
             // address that sits where we wnat by adding in the offset
             // from the beginning of the section.
             //


             ulMemoryBaseEx = (ULONG) virtualAddress;

             ulMemoryBaseEx += (ULONG)tmpPhysicalAddressBase.LowPart - (ULONG)viewBase.LowPart;

             debugOut(dbTrace,"mapped to %X",ulMemoryBaseEx);
            }
           }
         }
       }
     }

    ZwClose (physicalMemoryHandle);
   }


 #endif


  return ulMemoryBaseEx;
 }


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
NTSTATUS
CIOAccessDevice::allocMemory(
                             ULONG    ulLength,
                             PULONG   pulUserAddress,
                             PULONG   pulSystemAddress
                            )
 {
  NTSTATUS      ntStatus;
  PMemoryNode   node;
  ULONG         ulIndex;

  //
  //  First alloc our own free memory descriptor
  //

  for ( ulIndex = 0; ulIndex < MAX_FREE_MEMORY_NODES; ulIndex++)
   {
    node = &memoryList[ ulIndex ];

    if ( ! node->ulSystemAddress ) break;
   }

  if ( ulIndex >= MAX_FREE_MEMORY_NODES)
   {
    debugOut(dbError," ! no free memory descriptor available");
    return STATUS_INSUFFICIENT_RESOURCES;
   }

  ntStatus = STATUS_SUCCESS;

  #if defined (_NTKERNEL_)

   PHYSICAL_ADDRESS highestAcceptableAddress;

   highestAcceptableAddress.LowPart  =  -1;
   highestAcceptableAddress.HighPart =  -1;

   node->ulSystemAddress = (ULONG) MmAllocateContiguousMemory(ulLength,highestAcceptableAddress);

   if ( node->ulSystemAddress )
    {
     debugOut(dbTrace,"alloc %lu bytes of system memory %X",ulLength,node->ulSystemAddress);

     node->mdl = IoAllocateMdl((PVOID) node->ulSystemAddress,ulLength,FALSE,FALSE,NULL);
     if ( ! node->mdl )
      {
       MmFreeContiguousMemory( (PVOID) node->ulSystemAddress );
       node->ulSystemAddress = 0;
       debugOut(dbError,"! cannot alloc MDL");
       ntStatus = STATUS_INSUFFICIENT_RESOURCES;
      }
     else
      {
       //
       // build the MDL to desribe the memory pages
       //

       MmProbeAndLockPages( node->mdl, KernelMode, IoModifyAccess );

       //
       // Map locked pages into process's user address space
       //

       node->ulUserAddress = (ULONG) MmMapLockedPages( node->mdl, UserMode );
      }
    }
  #else

  #endif


  *pulUserAddress     = node->ulUserAddress;
  *pulSystemAddress   = node->ulSystemAddress;

  return ntStatus;
 }



//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
NTSTATUS
CIOAccessDevice::freeMemory(ULONG ulUserAddress)
 {
  PMemoryNode node;
  ULONG       ulIndex;

  //
  // First search memory node with given ulUserAddress
  //

  debugOut(dbTrace,"freeMemory() user address %X",ulUserAddress);

  for ( ulIndex = 0; ulIndex < MAX_FREE_MEMORY_NODES; ulIndex++)
   {
    node = &memoryList[ ulIndex ];

    if ( node->ulUserAddress == ulUserAddress )
     {
      debugOut(dbTrace,"freeMemory user address %X system address %X",ulUserAddress,node->ulSystemAddress);

      freeMemory ( node );

      return STATUS_SUCCESS;
     }
   }

  debugOut(dbError," ! memory address not found");
  return STATUS_ACCESS_DENIED;
 }


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
CIOAccessDevice::freeMemory(PMemoryNode node )
 {
  #if defined (_NTKERNEL_)

   MmUnmapLockedPages( (PVOID) node->ulUserAddress, node->mdl );

   IoFreeMdl( node->mdl );

   MmFreeContiguousMemory( (PVOID) node->ulSystemAddress );

  #else

  #endif

  node->ulSystemAddress = 0;
  node->ulUserAddress   = 0;
  node->mdl             = 0;
 }




