//---------------------------------------------------------------------------
// FileName     : pcienum.cpp
// Created      : 19/11/98
// Project      : Windows kernel device driver for hardware IO
// Author       : Elli
// Note         : This program is free software; you can redistribute it
//                and/or modify it under the terms of the GNU General Public
//                License as published by the Free Software Foundation;
//                either version 2 of the License, or (at your option) any
//                later version.
//
//                This program is distributed in the hope that it will be useful,
//                but WITHOUT ANY WARRANTY; without even the implied warranty of
//                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//                GNU General Public License for more details.
//
//                You should have received a copy of the GNU General Public
//                License along with this program; if not, write to the Free
//                Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// History      :
//                V 1.00 initial version
//---------------------------------------------------------------------------

#include "ioclass.h"
#include <debugout.h>
#include <stdio.h>
#include <string.h>

#if defined (_NTKERNEL_)
//---------------------------------------------------------------------------
// Find the PCI devices for the VendorId and DeviceId specified by the
// caller.
//---------------------------------------------------------------------------
NTSTATUS
CIOAccessDevice::pciFindDevice(
                               IN   USHORT vendorID, // The manufacturer of the device.  Assigned by the PCI SIG.
                               IN   USHORT deviceID, // The particular device created by the manufacturer.  Assigned by
                                                     // the manufacturer (via the PCI SIG).
                               OUT  PULONG  pdwBusNumber,
                               OUT  PULONG  pdwSlotNumber
                              )
 {
  PCI_SLOT_NUMBER         slotNumber;
  PCI_COMMON_CONFIG       pciData;
  ULONG                   busNumber;
  ULONG                   deviceNumber;
  ULONG                   functionNumber;

  // Remeber ...
  //
  // typedef struct _PCI_SLOT_NUMBER {
  //     union {
  //         struct {
  //             ULONG   DeviceNumber:5;
  //             ULONG   FunctionNumber:3;
  //             ULONG   Reserved:24;
  //         } bits;
  //         ULONG   AsULONG;
  //     } u;
  // } PCI_SLOT_NUMBER, *PPCI_SLOT_NUMBER;
  //

  slotNumber.u.AsULONG = 0;

  //
  // Scan each bus.
  //

  for (busNumber = 0; busNumber < 256; busNumber++)
   {
    //
    // Scan each device.
    //

    for (deviceNumber = 0; deviceNumber < PCI_MAX_DEVICES; deviceNumber++)
     {
      slotNumber.u.bits.DeviceNumber = deviceNumber;

      //
      // Scan each function.
      //

      for (functionNumber = 0; functionNumber < PCI_MAX_FUNCTION; functionNumber++)
       {
        slotNumber.u.bits.FunctionNumber = functionNumber;

        //
        // Check what's in the current slot.
        // Read the complete 256 bytes of configuration information
        //

        if ( HalGetBusData(
                             PCIConfiguration,
                             busNumber,
                             slotNumber.u.AsULONG,
                             &pciData,
                             sizeof(ULONG)
                            ) == 0)
         {
          //
          // The specified PCI bus does not exist.  We are done
          // with this bus.  Set the device number to the maximum
          // and break out of the function loop.  This will scan
          // the next system bus.  There is no guarantee that the
          // system buses are sequentially ordered.
          //

          deviceNumber = PCI_MAX_DEVICES;
          break;
         }

        if ( pciData.VendorID != 0xFFFF)
         {
          debugOut(dbTrace,"found pci device %lX %lX",pciData.VendorID,pciData.DeviceID);
         }

        if ( (pciData.VendorID == vendorID) && (pciData.DeviceID == deviceID) )
         {
          //
          // At this point, we've found a valid PCI device.
          // Get back the information for the located device.
          //

          *pdwBusNumber  = busNumber;
          *pdwSlotNumber = slotNumber.u.AsULONG;

          return STATUS_SUCCESS;
         }
       } // functionNumber

     }   // deviceNumber

   }   // busNumber

  return STATUS_DEVICE_DOES_NOT_EXIST;
 }


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
NTSTATUS
CIOAccessDevice::pciGetDeviceConfig(
                                    IN   ULONG              dwBusNumber,
                                    IN   ULONG              dwSlotNumber,
                                    OUT  PPCI_COMMON_CONFIG pciConfiguration
                                   )
 {
  NTSTATUS status;

  debugOut(dbTrace,"get pci configuration for bus %X slot %X",dwBusNumber,dwSlotNumber);

  status = HalGetBusData(
                          PCIConfiguration,
                          dwBusNumber,
                          dwSlotNumber,
                          pciConfiguration,
                          sizeof(*pciConfiguration)
                         );

  #if defined (DEBUG)
   debugOut(dbTrace,"found address %lX, irq %lX",
              pciConfiguration->u.type0.BaseAddresses[0],
              pciConfiguration->u.type0.InterruptLine   );
  #endif

  return STATUS_SUCCESS;
 }

#else


//---------------------------------------------------------------------------
// We can find some PCI information for WINDOWS 9x in the registry
//---------------------------------------------------------------------------
NTSTATUS
CIOAccessDevice::pciFindDevice(
                               IN  USHORT  vendorID,
                               IN  USHORT  deviceID,
                               OUT PULONG  pdwBusNumber,
                               OUT PULONG  pdwSlotNumber
                              )
 {
  DEVNODE               devNode;
  CHAR                  registryKeyPath[128];
  CHAR                  registrySubkeyPath[128];
  CHAR                  hardwareID[128];
  HKEY                  registryKey;
  ULONG                 status;
  int                   iIndex;
  CONFIGRET             cr;
  int                   bFound;

  sprintf(registryKeyPath, "VEN_%04X&DEV_%04X", vendorID, deviceID);

  debugOut(dbTrace,"search for <%s>",registryKeyPath);

  //
  // Typically this is the key, since some revision information are avail we must
  // enum all subkeys from \HKL\ENUM\PCI and compare with given registryKeyPath
  //

  *pdwBusNumber  = 0;
  *pdwSlotNumber = 0;
  bFound         = 0;

  status = _RegOpenKey(HKEY_LOCAL_MACHINE,"ENUM\\PCI",&registryKey);
  if ( status != STATUS_SUCCESS)
   {
    debugOut(dbError,"! cannot open HKL\\ENUM\\PCI error %X",status);
    return status;
   }

  for ( iIndex = 0; ! bFound && (status == STATUS_SUCCESS); iIndex++)
   {
    status = _RegEnumKey(
                          registryKey,
                          iIndex,
                          registrySubkeyPath,
                          sizeof(registrySubkeyPath)
                        );

    if ( status == STATUS_SUCCESS)
     {
       debugOut(dbTrace," found <%s>", registrySubkeyPath);


       bFound = (strstr(registrySubkeyPath,  registryKeyPath ) != NULL);
     }
   }

  _RegCloseKey(registryKey);


  if ( bFound )
   {
    //
    // Now we have the full path name for given vendor ID
    //

    sprintf( registryKeyPath, "ENUM\\PCI\\%s",registrySubkeyPath );

    debugOut(dbTrace," found pci board %s, get configuration",registryKeyPath);

    status = _RegOpenKey(HKEY_LOCAL_MACHINE, registryKeyPath, &registryKey );
    if ( status == STATUS_SUCCESS)
     {
      debugOut(dbTrace,"found hardware key <%s>",registryKeyPath );

      bFound = FALSE;

      for ( iIndex = 0; ! bFound && (status == STATUS_SUCCESS); iIndex++)
       {
        status = _RegEnumKey(
                            registryKey,
                            iIndex,
                            registrySubkeyPath,
                            sizeof(registrySubkeyPath)
                            );

        if ( status == STATUS_SUCCESS)
         {
          debugOut(dbTrace," subkey  <%s>", registrySubkeyPath);

          sprintf(hardwareID,"%s\\%s",&registryKeyPath[5],registrySubkeyPath);

          debugOut(dbTrace," locate device node for <%s>",hardwareID);

          cr      = _CONFIGMG_Locate_DevNode((DEVNODE *)pdwSlotNumber, hardwareID, 0);
          if (cr == STATUS_SUCCESS)
           {
            bFound = TRUE;

            debugOut(dbTrace," device node %X found", *pdwSlotNumber);
           }
         }
       }
      _RegCloseKey(registryKey);
     }
    else
     {
      debugOut(dbError,"! cannot open key <%s>",registryKeyPath);
     }


   }  // root key found

  if ( *pdwSlotNumber == 0)
   {
    debugOut(dbError," key not found, PCI board not installed");
    return ERROR_ACCESS_DENIED;
   }

  return STATUS_SUCCESS;
 }


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
NTSTATUS
CIOAccessDevice::pciGetDeviceConfig(
                                    IN   ULONG              dwBusNumber,
                                    IN   ULONG              dwSlotNumber,
                                    OUT  PPCI_COMMON_CONFIG pciConfiguration
                                   )
 {
  CONFIGRET cr;

  cr = _CONFIGMG_Call_Enumerator_Function(
                                          dwSlotNumber,
                                          PCI_ENUM_FUNC_GET_DEVICE_INFO,
                                          0,
                                          pciConfiguration,
                                          sizeof(*pciConfiguration),
                                          0
                                        );


  debugOut(dbTrace,"get PCI device info for node %lX returns %X",dwSlotNumber,cr);

  return cr;
 }
#endif






